/*
 * File: bsd/parmacs_support.c
 * Description: implementation of support code for BSD based PARMACS
 *   macros. Also define global var definitions.
 * Comments: 
 * Copyright: Department of Computer Architecture - Polytechnic University of
 *   Catalonia (UPC)
 * Created: 1999/06 ernest@ac.upc.es (E.Artiaga)
 * Contact: gsomk@ac.upc.es
 * Last update:
 * History:
 */
/*
 * Disclaimer and copying policy:
 *
 * The Polytechnic University of Catalonia (UPC) and the authors disclaim
 * all warranties with regard to this software, including all implied
 * warranties of merchantability and fitness.
 * In no event shall the Polytechnic University of Catalonia (UPC) or the
 * authors be liable for any special, indirect or consequential damages or
 * any damages whatsoever resulting from loss of use, data or profits,
 * whether in an action of contract, negligence or other tortius action,
 * arising out of or in connection with the use or performance of this
 * software.
 *
 * Permission to use, copy, modify, and distribute this software for
 * any purpose and without fee is hereby granted, provided that the above
 * copyright notice and disclaimer, and also this permission notice, appear
 * in all copies.
 *
 */

#include <parmacs_internals.h>

#ifndef MAP_ANONYMOUS
#ifdef MAP_ANON
#define MAP_ANONYMOUS MAP_ANON
#endif
#endif

#ifdef MAP_ANONYMOUS
#ifdef MAP_VARIABLE
#define MAP_FLAGS MAP_ANONYMOUS|MAP_SHARED|MAP_VARIABLE
#else /* MAP_VARIABLE */
#define MAP_FLAGS MAP_ANONYMOUS|MAP_SHARED
#endif /* MAP_VARIABLE */
#else /* MAP_ANONYMOUS */
#ifdef MAP_VARIABLE
#define MAP_FLAGS MAP_SHARED|MAP_VARIABLE
#else /* MAP_VARIABLE */
#define MAP_FLAGS MAP_SHARED
#endif /* MAP_VARIABLE */
#endif /* MAP_ANONYMOUS */

/*
 * Data structures
 */
 
parmacs_data_t _parmacs_data = (parmacs_data_t)0;
parmacs_configuration _parmacs_configuration = {
  (char *)0,
  PARMACS_MMAP_DEVICE,
  PARMACS_BUSY_WAIT_COUNT,
  PARMACS_SPIN_COUNT,
  PARMACS_MAX_BARRIER,
  PARMACS_MAX_GLOBAL,
  PARMACS_MAX_EVENT,
  PARMACS_MAX_LOCK,
  PARMACS_PAGE_SIZE,
  PARMACS_STACK_SIZE,
  PARMACS_SHARED_MEMORY,
  PARMACS_MEM_ALIGNMENT,
  PARMACS_MAX_PROCESS,
  PARMACS_TIMESTAMP
};
parmacs_configuration_t _parmacs_config = &_parmacs_configuration;

parmacs_process_t _parmacs_self = (parmacs_process_t)0;

static void _parmacs_configure(unsigned long int);

/*
 * Implement msemaphores if needed
 */
#ifdef PARMACS_NO_MSEM

msemaphore *
msem_init(msemaphore *sem, int init_value)
{
    sem->lock = init_value;
    return(sem);
}

int 
msem_lock(msemaphore *sem, int cond)
{
    register int countdown;
    register int spin = _parmacs_config->spin_count;
    
    int value = parmacs_test_and_set(&(sem->lock), MSEM_LOCKED);
    if (value == MSEM_UNLOCKED) return(0);
    if (cond) return(-1);
    countdown = spin;
    while(parmacs_test_and_set(&(sem->lock), MSEM_LOCKED) != MSEM_UNLOCKED) {
      while(sem->lock == MSEM_LOCKED) {
        if (countdown >= 0) {
          if (!countdown) {
            _parmacs_yield();
            countdown = spin;
          } else {
            countdown --;
          }
        }
      }
    }
    return 0;
}

int 
msem_unlock(msemaphore *sem, int cond)
{
    cond = cond;
    sem->lock = MSEM_UNLOCKED;
    return(0);
}

#endif /* PARMACS_NO_MSEM */

/*
 * Initialize environment
 * Comments:
 *   This routine is supposed to be called from the main routine, at the
 *   very beginning of the execution.
 */
void
_parmacs_main_initenv(unsigned long int shm_size)
{
  register void *address;
  int mmap_fd;
  register int i;
#ifdef PARMACS_NO_SCHED_YIELD
  int blocking_device[2];
#endif /* PARMACS_NO_SCHED_YIELD */
  
  /* Load configuration */
  _parmacs_configure(shm_size);

  /* Prepare file descriptor for mmap */
  if (_parmacs_config->mmap_device && *_parmacs_config->mmap_device) {
    if ((mmap_fd=open(_parmacs_config->mmap_device,O_RDWR))<0){
      _parmacs_panic("parmacs: cannot open device for anonymous mmap!");
    }
  } else {
    mmap_fd = -1;
  } 
   
  /* Allocate memory for parmacs data */
  /* MAIN DATA */
  address = mmap((void *)0, sizeof(parmacs_data),
                 PROT_READ|PROT_WRITE, MAP_FLAGS, mmap_fd, 0);
  if ((int)address == -1) {
    _parmacs_panic("parmacs: cannot allocate memory for parmacs data!");
  }
  _parmacs_data = (parmacs_data_t)address;
  _parmacs_data->mmap_fd = mmap_fd;
  
  /* Initialize mutex locks */
  address = mmap((void *)0, sizeof(msemaphore)*PARMACS_MAX_ID,
                 PROT_READ|PROT_WRITE, MAP_FLAGS, mmap_fd, 0);
  if ((int)address == -1) {
    _parmacs_panic("parmacs: cannot allocate semaphores for mutual exclusion!");
  }
  _parmacs_data->mutex = address;
  for (i=0; i < PARMACS_MAX_ID; i++) {
    if (msem_init(&(_parmacs_data->mutex[i]), MSEM_UNLOCKED) == NULL) {
       _parmacs_panic("parmacs: cannot initialize semaphores for mutex!");
    }
  }
  
  /* MEMORY */
  address = mmap((void *)0, _parmacs_config->shared_memory,
                 PROT_READ|PROT_WRITE, MAP_FLAGS, mmap_fd, 0);
  if ((int)address == -1) {
    _parmacs_panic("parmacs: cannot allocate shared memory!");
  }
  _parmacs_data->shm_address = address;
  _parmacs_data->shm_pointer = address;
  _parmacs_data->shm_size    = _parmacs_config->shared_memory;
  _parmacs_data->shm_avail   = _parmacs_config->shared_memory;
  
  /* BARRIERS */
  address = mmap((void *)0,
                 _parmacs_config->max_barriers*sizeof(parmacs_barrier),
                 PROT_READ|PROT_WRITE, MAP_FLAGS, mmap_fd, 0);
  if ((int)address == -1) {
    _parmacs_panic("parmacs: cannot allocate memory for barriers!");
  }
  _parmacs_data->barrier = address;
  _parmacs_data->n_barriers = 0;
  
  /* GLOBAL SUBSCRIPTS */
  address = mmap((void *)0,
                 _parmacs_config->max_globals*sizeof(parmacs_global),
                 PROT_READ|PROT_WRITE, MAP_FLAGS, mmap_fd, 0);
  if ((int)address == -1) {
    _parmacs_panic("parmacs: cannot allocate memory for global subscripts!");
  }
  _parmacs_data->global = address;
  _parmacs_data->n_globals = 0;
  
  /* EVENTS */
  address = mmap((void *)0,
                 _parmacs_config->max_events*sizeof(parmacs_event),
                 PROT_READ|PROT_WRITE, MAP_FLAGS, mmap_fd, 0);
  if ((int)address == -1) {
    _parmacs_panic("parmacs: cannot allocate memory for events!");
  }
  _parmacs_data->event = address;
  _parmacs_data->n_events = 0;
  
  /* LOCKS */
  address = mmap((void *)0,
                 _parmacs_config->max_locks*sizeof(parmacs_lock),
                 PROT_READ|PROT_WRITE, MAP_FLAGS, mmap_fd, 0);
  if ((int)address == -1) {
    _parmacs_panic("parmacs: cannot allocate memory for locks!");
  }
  _parmacs_data->lock = address;
  _parmacs_data->n_locks = 0;
  
  /* PROCESSES */
  address = mmap((void *)0,
                 _parmacs_config->max_processes*sizeof(parmacs_process),
                 PROT_READ|PROT_WRITE, MAP_FLAGS, mmap_fd, 0);
  if ((int)address == -1) {
    _parmacs_panic("parmacs: cannot allocate memory for processes!");
  }
  _parmacs_data->process = address;
  for(i=1; i < (_parmacs_config->max_processes-1); i++) {
    _parmacs_data->process[i].link = &(_parmacs_data->process[i+1]);
  }
  _parmacs_data->process[_parmacs_config->max_processes-1].link =
      (parmacs_process_t)0;
  _parmacs_data->free_process = &(_parmacs_data->process[1]);
  _parmacs_self = &(_parmacs_data->process[0]);
  _parmacs_self->event = (parmacs_event_t)0;
  
#ifdef PARMACS_NO_SCHED_YIELD  
  /*
   * Initialize blocking device to implement processor yields
   */
  if (pipe(blocking_device) < 0) {
    _parmacs_panic("parmacs: cannot create pipe for yielding!");
  }
  _parmacs_data->blocking_fd = blocking_device[0];
#endif /* PARMACS_NO_SCHED_YIELD */

  /*
   * Time stamp:
   */
  if (_parmacs_config->timestamp) {
    struct tms now;
    _parmacs_data->start_time = times(&now);
  }
}

void
_parmacs_main_end()
{
  /*
   * Time stamp:
   */
  if (_parmacs_config->timestamp) {
    struct tms now;
    _parmacs_data->end_time = times(&now);
    fprintf(stderr, "\nPARMACS TIME: start: %lu, end: %lu, elapsed: %lu.\n",
            _parmacs_data->start_time,
            _parmacs_data->end_time,
            _parmacs_data->end_time - _parmacs_data->start_time);
  }
  _parmacs_exit();
}

static void
_parmacs_configure(unsigned long int shm_size)
{
  extern char **environ;
  char **my_environment;
  char * var_name;

  _parmacs_configuration.shared_memory = shm_size;
  
  /* Read configuration from environment */
  
  for(my_environment = environ; *my_environment; my_environment++) {
    var_name = strtok(strdup(*my_environment), "=");
    if (!strcmp(var_name, "PARMACS_CONFIG_FILE")) {
      _parmacs_configuration.config_file = strdup(strtok(NULL, "\0"));
    } else if (!strcmp(var_name, "PARMACS_MMAP_DEVICE")) {
      _parmacs_configuration.mmap_device = strdup(strtok(NULL, "\0"));
    } else if (!strcmp(var_name, "PARMACS_BUSY_WAIT_COUNT")) {
      _parmacs_configuration.busy_wait_count = atoi(strtok(NULL, "\0"));
    } else if (!strcmp(var_name, "PARMACS_SPIN_COUNT")) {
      _parmacs_configuration.spin_count = atoi(strtok(NULL, "\0"));
    } else if (!strcmp(var_name, "PARMACS_MAX_BARRIER")) {
      _parmacs_configuration.max_barriers = atoi(strtok(NULL, "\0"));
    } else if (!strcmp(var_name, "PARMACS_MAX_GLOBAL")) {
      _parmacs_configuration.max_globals = atoi(strtok(NULL, "\0"));
    } else if (!strcmp(var_name, "PARMACS_MAX_EVENT")) {
      _parmacs_configuration.max_events = atoi(strtok(NULL, "\0"));
    } else if (!strcmp(var_name, "PARMACS_MAX_LOCK")) {
      _parmacs_configuration.max_locks = atoi(strtok(NULL, "\0"));
    } else if (!strcmp(var_name, "PARMACS_PAGE_SIZE")) {
      _parmacs_configuration.page_size = atoi(strtok(NULL, "\0"));
    } else if (!strcmp(var_name, "PARMACS_STACK_SIZE")) {
      _parmacs_configuration.stack_size = atoi(strtok(NULL, "\0"));
    } else if (!strcmp(var_name, "PARMACS_SHARED_MEMORY")) {
      char *endptr;
      _parmacs_configuration.shared_memory = 
        strtoul(strtok(NULL, "\0"), &endptr, 10);
    } else if (!strcmp(var_name, "PARMACS_MEM_ALIGNMENT")) {
      _parmacs_configuration.memory_alignment = atoi(strtok(NULL, "\0"));
    } else if (!strcmp(var_name, "PARMACS_MAX_PROCESS")) {
      _parmacs_configuration.max_processes = atoi(strtok(NULL, "\0"));
    } else if (!strcmp(var_name, "PARMACS_TIMESTAMP")) {
      _parmacs_configuration.timestamp = atoi(strtok(NULL, "\0"));
    }
    free(var_name);  
  }
  
  /* if there is a configuration file, use it */
  if (_parmacs_configuration.config_file) {
    char buffer[512];
    FILE *fd;

    if (!(fd = fopen(_parmacs_configuration.config_file, "r"))) {
      _parmacs_panic("parmacs: cannot open configuration file!");
    }    
    while (fgets(buffer, 512, fd)) {
      var_name = strtok(buffer, " =\t\n\0");
      if (!strcmp(var_name, "PARMACS_MMAP_DEVICE")) {
        _parmacs_configuration.mmap_device = strdup(strtok(NULL, " =\t\n\0"));
      } else if (!strcmp(var_name, "PARMACS_BUSY_WAIT_COUNT")) {
        _parmacs_configuration.busy_wait_count = atoi(strtok(NULL, " =\t\n\0"));
      } else if (!strcmp(var_name, "PARMACS_SPIN_COUNT")) {
        _parmacs_configuration.spin_count = atoi(strtok(NULL, " =\t\n\0"));
      } else if (!strcmp(var_name, "PARMACS_MAX_BARRIER")) {
        _parmacs_configuration.max_barriers = atoi(strtok(NULL, " =\t\n\0"));
      } else if (!strcmp(var_name, "PARMACS_MAX_GLOBAL")) {
        _parmacs_configuration.max_globals = atoi(strtok(NULL, " =\t\n\0"));
      } else if (!strcmp(var_name, "PARMACS_MAX_EVENT")) {
        _parmacs_configuration.max_events = atoi(strtok(NULL, " =\t\n\0"));
      } else if (!strcmp(var_name, "PARMACS_MAX_LOCK")) {
        _parmacs_configuration.max_locks = atoi(strtok(NULL, " =\t\n\0"));
      } else if (!strcmp(var_name, "PARMACS_PAGE_SIZE")) {
        _parmacs_configuration.page_size = atoi(strtok(NULL, " =\t\n\0"));
      } else if (!strcmp(var_name, "PARMACS_STACK_SIZE")) {
        _parmacs_configuration.stack_size = atoi(strtok(NULL, " =\t\n\0"));
      } else if (!strcmp(var_name, "PARMACS_SHARED_MEMORY")) {
        char *endptr;
        _parmacs_configuration.shared_memory = 
          strtoul(strtok(NULL, " =\t\n\0"), &endptr, 10);
      } else if (!strcmp(var_name, "PARMACS_MEM_ALIGNMENT")) {
        _parmacs_configuration.memory_alignment=atoi(strtok(NULL, " =\t\n\0"));
      } else if (!strcmp(var_name, "PARMACS_MAX_PROCESS")) {
        _parmacs_configuration.max_processes = atoi(strtok(NULL, " =\t\n\0"));
      } else if (!strcmp(var_name, "PARMACS_TIMESTAMP")) {
        _parmacs_configuration.timestamp = atoi(strtok(NULL, " =\t\n\0"));
      }
    }
    fclose(fd);
  }
}

/*
 * Memory management
 * Comments:
 *   Memory management is extremely simple in this implementation. The
 *   allocation routine simply assigns consecutive regions within the
 *   shared memory area, till that area is full. Freeing memory is not
 *   implemented, so that memory can never be reused.
 *   The allocation routine returns an address which is properly aligned,
 *   or zero if there is not enough memory.
 */
 
void *
_parmacs_memory_alloc(int size)
{
  register void *address;
  register int vsize = size;
  register unsigned long int u_ptr;
  register int adjust;
  
  _parmacs_internal_lock(PARMACS_MEMORY_ID);
  u_ptr = (unsigned long int)_parmacs_data->shm_pointer;
  if ((adjust = u_ptr & (_parmacs_config->memory_alignment-1))) {
    u_ptr += _parmacs_config->memory_alignment - adjust;
    vsize += _parmacs_config->memory_alignment - adjust;
  }
  if ((adjust = size & (_parmacs_config->memory_alignment-1))) {
    size  += _parmacs_config->memory_alignment - adjust;
    vsize += _parmacs_config->memory_alignment - adjust;
  }
  if (vsize > _parmacs_data->shm_avail) {
    _parmacs_internal_unlock(PARMACS_MEMORY_ID);
    return (void *)0;
  }
  address = (void *)u_ptr;
  _parmacs_data->shm_pointer = (void *)(u_ptr + size);
  _parmacs_data->shm_avail -= vsize;
  _parmacs_internal_unlock(PARMACS_MEMORY_ID);
  return (void *)address;
}

void
_parmacs_memory_free(void *address)
{
  /* not implemented */
}

void *
_parmacs_local_alloc(int size)
{
  void *address;
  _parmacs_internal_lock(PARMACS_MEMORY_ID);
  address = (void *)malloc(size);
  _parmacs_internal_unlock(PARMACS_MEMORY_ID);
  return address;
}

void
_parmacs_local_free(void *address)
{
  _parmacs_internal_lock(PARMACS_MEMORY_ID);
  free(address);
  _parmacs_internal_unlock(PARMACS_MEMORY_ID);
}

/*
 * Process creation and joining
 */
 
void
_parmacs_create(void (*routine)())
{
  register int i;
  parmacs_process_t child;

  /* Allocate an entry for the new process */  
  _parmacs_internal_lock(PARMACS_PROCESS_ID);
  if (!_parmacs_data->free_process) {
    _parmacs_internal_unlock(PARMACS_PROCESS_ID);
    _parmacs_panic("parmacs: too many processes!\n");
  }
  child = _parmacs_data->free_process;
  child->event = (parmacs_event_t)0;
  _parmacs_data->free_process = child->link;
  child->link = (parmacs_process_t)0;
  _parmacs_internal_unlock(PARMACS_PROCESS_ID);

  /* Actually fork */
  switch(fork()) {
  case -1:
    _parmacs_panic("parmacs: cannot create new process!");
  case 0:
    /* self identify */
    _parmacs_self = child;
    /* execute code */
    (*routine)();
    /* clean up structures */
    _parmacs_internal_lock(PARMACS_PROCESS_ID);
    _parmacs_self->link = _parmacs_data->free_process;
    _parmacs_data->free_process = _parmacs_self;
    _parmacs_internal_unlock(PARMACS_PROCESS_ID);
    exit(0);
  default:
    break;
  }
}

void
_parmacs_wait_for_end(int amount)
{
  register int i;
  int status;
  
  for(i = 0; i < amount; i++) {
    if (wait(&status) == -1 && errno == EINTR) i--;
  }
}

parmacs_pid
_parmacs_get_pid()
{
  return ((parmacs_pid)getpid());
}

/*
 * Scheduling, clean up and exit routines
 */
 
void 
_parmacs_panic(char *msg)
{
  if (msg) fprintf(stderr, "%s\n", msg);
  kill(0, SIGKILL);
  exit(-1);
}

void 
_parmacs_exit()
{
  exit(0);
}

void 
_parmacs_yield()
{
#ifndef PARMACS_NO_SCHED_YIELD
  sched_yield();
#else /* PARMACS_NO_SCHED_YIELD */
  struct timeval t;
  fd_set fds1, fds2, fds3;
  t.tv_sec  = 0;
  t.tv_usec = 1;
  FD_ZERO(&fds1);
  FD_ZERO(&fds2);
  FD_ZERO(&fds3);
  FD_SET(_parmacs_data->blocking_fd, &fds1);
  select(1, &fds1, &fds2, &fds3, &t);
#endif /* PARMACS_NO_SCHED_YIELD */
}

/*
 * Clock
 */
parmacs_timestamp
_parmacs_clock()
{
  struct tms now;
  return (parmacs_timestamp)times(&now);
}

