/*
 * File: bsd/parmacs_sync.c
 * Description: implementation of synchronization macros of PARMACS
 * Comments: 
 * Copyright: Department of Computer Architecture - Polytechnic University of
 *   Catalonia (UPC)
 * Created: 1999/06 ernest@ac.upc.es (E.Artiaga)
 * Contact: gsomk@ac.upc.es
 * Last update:
 * History:
 */
/*
 * Disclaimer and copying policy:
 *
 * The Polytechnic University of Catalonia (UPC) and the authors disclaim
 * all warranties with regard to this software, including all implied
 * warranties of merchantability and fitness.
 * In no event shall the Polytechnic University of Catalonia (UPC) or the
 * authors be liable for any special, indirect or consequential damages or
 * any damages whatsoever resulting from loss of use, data or profits,
 * whether in an action of contract, negligence or other tortius action,
 * arising out of or in connection with the use or performance of this
 * software.
 *
 * Permission to use, copy, modify, and distribute this software for
 * any purpose and without fee is hereby granted, provided that the above
 * copyright notice and disclaimer, and also this permission notice, appear
 * in all copies.
 *
 */

#include <parmacs_internals.h>


/*
 * Barriers
 */

void
_parmacs_barinit(parmacs_barrier_t *barrier)
{
  
  /* get a free barrier entry */
  _parmacs_internal_lock(PARMACS_BARRIER_ID);
  if (_parmacs_data->n_barriers >= _parmacs_config->max_barriers) {
    _parmacs_internal_unlock(PARMACS_BARRIER_ID);
    _parmacs_panic("parmacs: too many barriers!");
  }
  *barrier = &_parmacs_data->barrier[_parmacs_data->n_barriers++];
  _parmacs_internal_unlock(PARMACS_BARRIER_ID);
  
  /* now, initialize the new barrier */
  if (msem_init(&((*barrier)->s), MSEM_UNLOCKED) == NULL) {
    _parmacs_panic("parmacs: cannot initialize semaphore for barrier!");
  }
  (*barrier)->u.data.n = 0;
  (*barrier)->u.data.seq = 0;
  (*barrier)->u.data.active = 0;
}
 
void 
_parmacs_barrier(parmacs_barrier_t barrier, int procs)
{
  register int countdown;
  register int spin = _parmacs_config->busy_wait_count;
    
  /* Enter the barrier */
  _parmacs_sem_lock(&(barrier->s));
  if (!barrier->u.data.active) {
    /* Setup the barrier */
    barrier->u.data.active = 1;
    barrier->u.data.seq++;
  }
  
  barrier->u.data.n++;
  _parmacs_self->barrier_seq = barrier->u.data.seq;
  
  if (barrier->u.data.n == procs) {
    /* This is the last process - reset the barrier and wakeup everybody */
    barrier->u.data.n = 0;
    barrier->u.data.active = 0;
  }
  _parmacs_sem_unlock(&(barrier->s));
  
  /* now, wait for everybody to finish, and then go on */
  countdown = spin;
  while (barrier->u.data.active && 
         barrier->u.data.seq == _parmacs_self->barrier_seq) {
    if (countdown >= 0) {
      if (!countdown) {
        _parmacs_yield();
        countdown = spin;
      } else {
        countdown --;
      }
    }
  }
}

/*
 * Global subscripts
 */
 
void
_parmacs_gsinit(parmacs_global_t *global)
{
  
  /* get a free global subscript entry */
  _parmacs_internal_lock(PARMACS_GLOBAL_ID);
  if (_parmacs_data->n_globals >= _parmacs_config->max_globals) {
    _parmacs_internal_unlock(PARMACS_GLOBAL_ID);
    _parmacs_panic("parmacs: too many global subscripts!");
  }
  *global = &_parmacs_data->global[_parmacs_data->n_globals++];
  _parmacs_internal_unlock(PARMACS_GLOBAL_ID);
  
  /* now, initialize the new global subscript */
  if (msem_init(&((*global)->s), MSEM_UNLOCKED) == NULL) {
   _parmacs_panic("parmacs: cannot initialize semaphore for global subscript!");
  }
  (*global)->u.data.i = 0;
  (*global)->u.data.n = 0;
  (*global)->u.data.seq = 0;
  (*global)->u.data.active = 0;
}

void
_parmacs_getsub(parmacs_global_t global, int *next, int last, int procs)
{
  register int countdown;
  register int spin = _parmacs_config->busy_wait_count;
    
  _parmacs_sem_lock(&(global->s));
  
  if (!global->u.data.active) {
    /* First to enter: setup global subscript */
    global->u.data.active = 1;
    global->u.data.seq ++;
  }

  /* Get next subscript */
  if (global->u.data.i <= last) {
    *next = global->u.data.i++;
    _parmacs_sem_unlock(&(global->s));
  } else {
    *next = -1;
    /* the loop is finished: wait until everybody is done */
    global->u.data.n++;
    _parmacs_self->global_seq = global->u.data.seq;
    
    if (global->u.data.n == procs) {
      /* This is the last process - reset subscript and wakeup everybody */
      global->u.data.i = 0;
      global->u.data.n = 0;
      global->u.data.active = 0;
    }
    _parmacs_sem_unlock(&(global->s));

    /* now, wait for everybody to finish, and then go on */
    countdown = spin;
    while (global->u.data.active && 
           global->u.data.seq == _parmacs_self->global_seq) {
      if (countdown >= 0) {
        if (!countdown) {
          _parmacs_yield();
          countdown = spin;
        } else {
          countdown --;
        }
      }
    }
  }
}

/*
 * Locks
 */

void 
_parmacs_lockinit(parmacs_lock_t *lock)
{
  
  /* get a free lock entry */
  _parmacs_internal_lock(PARMACS_LOCK_ID);
  if (_parmacs_data->n_locks >= _parmacs_config->max_locks) {
    _parmacs_internal_unlock(PARMACS_LOCK_ID);
    _parmacs_panic("parmacs: too many locks!");
  }
  *lock = &_parmacs_data->lock[_parmacs_data->n_locks++];
  _parmacs_internal_unlock(PARMACS_LOCK_ID);
  
  /* now, initialize the new lock */
  if (msem_init(&((*lock)->s), MSEM_UNLOCKED) == NULL){
    _parmacs_panic("parmacs: cannot initialize semaphore for lock!");
  }
}

void
_parmacs_lock(parmacs_lock_t lock)
{
  _parmacs_sem_lock(&(lock->s));
}

void
_parmacs_unlock(parmacs_lock_t lock)
{
  _parmacs_sem_unlock(&(lock->s));
}

void
_parmacs_alockinit(parmacs_lock_t *lock, int amount)
{
  register int i;
  
  /* get a free lock entry */
  _parmacs_internal_lock(PARMACS_LOCK_ID);
  if ((_parmacs_data->n_locks + amount -1) >= _parmacs_config->max_locks) {
    _parmacs_internal_unlock(PARMACS_LOCK_ID);
    _parmacs_panic("parmacs: too many locks!");
  }
  *lock = &_parmacs_data->lock[_parmacs_data->n_locks];
  _parmacs_data->n_locks += amount;
  _parmacs_internal_unlock(PARMACS_LOCK_ID);
  
  /* now, initialize the new locks */
  for (i=0; i < amount; i++) {
    if (msem_init(&((*lock)[i].s), MSEM_UNLOCKED) == NULL) {
      _parmacs_panic("parmacs: cannot initialize semaphore for lock array!");
    }
  }
}

void
_parmacs_alock(parmacs_lock_t lock, int idx)
{
  _parmacs_sem_lock(&(lock[idx].s));
}

void
_parmacs_aulock(parmacs_lock_t lock, int idx)
{
  _parmacs_sem_unlock(&(lock[idx].s));
}

/*
 * Events
 */
 
void
_parmacs_pauseinit(parmacs_event_t *event, int amount)
{
  register int i;
  
  /* get the free event entries */
  _parmacs_internal_lock(PARMACS_EVENT_ID);
  if ((_parmacs_data->n_events + amount -1) >= _parmacs_config->max_events) {
    _parmacs_internal_unlock(PARMACS_EVENT_ID);
    _parmacs_panic("parmacs: too many events!");
  }
  *event = &_parmacs_data->event[_parmacs_data->n_events];
  _parmacs_data->n_events += amount;
  _parmacs_internal_unlock(PARMACS_EVENT_ID);
  
  /* initialize the event entries */
  for(i=0; i < amount; i++) {
    if (msem_init(&((*event)[i].s), MSEM_UNLOCKED) == NULL) {
      _parmacs_panic("parmacs: cannot initialize semaphore for event!");
    }
    (*event)[i].u.data.v = 0;
    (*event)[i].u.data.blocked = (void *)0;
  }
}

void
_parmacs_clearpause(parmacs_event_t event, int idx)
{
  /* must be locked to avoid interferences with other event operations */
  _parmacs_sem_lock(&(event[idx].s));
  event[idx].u.data.v = 0;
  _parmacs_sem_unlock(&(event[idx].s));
}

void
_parmacs_setpause(parmacs_event_t event, int idx)
{
  register int i;
  register parmacs_process_t process;
   
  _parmacs_sem_lock(&(event[idx].s));
  /* set the value */
  event[idx].u.data.v = 1;
  /* wakeup everybody waiting for this event */
  process = (parmacs_process_t)(event[idx].u.data.blocked);
  while (process) {
    event[idx].u.data.blocked = (void *)(process->link);
    process->link  = (parmacs_process_t)0;
    process->event = (parmacs_event_t)0;
    process = (parmacs_process_t)(event[idx].u.data.blocked);
  }
  _parmacs_sem_unlock(&(event[idx].s));
}

void
_parmacs_event(parmacs_event_t event, int idx)
{
  register int countdown;
  register int spin = _parmacs_config->busy_wait_count;
  register parmacs_process_t process;

  /* Wait for the event to be zero */  
  _parmacs_sem_lock(&(event[idx].s));
  while(event[idx].u.data.v == 1) {
    _parmacs_sem_unlock(&(event[idx].s));
    countdown = spin;
    while(event[idx].u.data.v == 1) {
      if (countdown >= 0) {
        if (!countdown) {
          _parmacs_yield();
          countdown = spin;
        } else {
          countdown --;
        }
      }
    }
    _parmacs_sem_lock(&(event[idx].s));
  }
  
  /* set the event again */
  event[idx].u.data.v = 1;    

  /* wakeup everybody waiting for this event to be set */
  process = (parmacs_process_t)(event[idx].u.data.blocked);
  while (process) {
    event[idx].u.data.blocked = (void *)(process->link);
    process->link  = (parmacs_process_t)0;
    process->event = (parmacs_event_t)0;
    process = (parmacs_process_t)(event[idx].u.data.blocked);
  }
  _parmacs_sem_unlock(&(event[idx].s));
}

void
_parmacs_waitpause(parmacs_event_t event, int idx)
{
  register int countdown;
  register int spin = _parmacs_config->busy_wait_count;
  
  /* check the event and prepare to block */
  _parmacs_sem_lock(&(event[idx].s));
  if (event[idx].u.data.v == 0) {
    _parmacs_self->link = (parmacs_process_t)(event[idx].u.data.blocked);
    event[idx].u.data.blocked = (void *)_parmacs_self;
    _parmacs_self->event = &event[idx];
  }
  _parmacs_sem_unlock(&(event[idx].s));
  
  /* now, wait */
  countdown = spin;
  while(_parmacs_self->event) {
    if (countdown >= 0) {
      if (!countdown) {
        _parmacs_yield();
        countdown = spin;
      } else {
        countdown --;
      }
    }
  }
}

void
_parmacs_pause(parmacs_event_t event, int idx)
{
  register int countdown;
  register int spin = _parmacs_config->busy_wait_count;

  /* Wait for the event to be set */  
  _parmacs_sem_lock(&(event[idx].s));
  while(event[idx].u.data.v == 0) {
    _parmacs_sem_unlock(&(event[idx].s));
    countdown = spin;
    while(event[idx].u.data.v == 0) {
      if (countdown >= 0) {
        if (!countdown) {
          _parmacs_yield();
          countdown = spin;
        } else {
          countdown --;
        }
      }
    }
    _parmacs_sem_lock(&(event[idx].s));
  }
  
  /* zero the event again */
  event[idx].u.data.v = 0;    
  _parmacs_sem_unlock(&(event[idx].s));
}  
  
  
