/*
 * File: bsd/parmacs_types.h
 * Description: miscellaneous type definitions
 * Comments: 
 * Copyright: Department of Computer Architecture - Polytechnic University of
 *   Catalonia (UPC)
 * Created: 1999/06 ernest@ac.upc.es (E.Artiaga)
 * Contact: gsomk@ac.upc.es
 * Last update:
 * History:
 */
/*
 * Disclaimer and copying policy:
 *
 * The Polytechnic University of Catalonia (UPC) and the authors disclaim
 * all warranties with regard to this software, including all implied
 * warranties of merchantability and fitness.
 * In no event shall the Polytechnic University of Catalonia (UPC) or the
 * authors be liable for any special, indirect or consequential damages or
 * any damages whatsoever resulting from loss of use, data or profits,
 * whether in an action of contract, negligence or other tortius action,
 * arising out of or in connection with the use or performance of this
 * software.
 *
 * Permission to use, copy, modify, and distribute this software for
 * any purpose and without fee is hereby granted, provided that the above
 * copyright notice and disclaimer, and also this permission notice, appear
 * in all copies.
 *
 */

#ifndef _BSD_PARMACS_TYPES_H_
#define _BSD_PARMACS_TYPES_H_

#include <stdlib.h>
#include <sys/mman.h>

#ifdef PARMACS_INTERNALS
/* Parmacs configuration */
typedef struct parmacs_configuration {
  char              *config_file;
  char              *mmap_device;
  int                busy_wait_count;
  int                spin_count;
  int                max_barriers;
  int                max_globals;
  int                max_events;
  int                max_locks;
  int                page_size;
  int                stack_size;
  unsigned long int  shared_memory;
  int                memory_alignment;
  int                max_processes;
  int                timestamp;
} parmacs_configuration;

typedef parmacs_configuration *parmacs_configuration_t;

#endif /* PARMACS_INTERNALS */

#ifdef PARMACS_NO_MSEM

typedef struct msemaphore {
  volatile int lock;
} msemaphore;

#endif /* PARMACS_NO_MSEM */

/* Some systems need msemaphores to be properly aligned. use MSEM_INT for pad */
#define MSEM_PAD(x) (((x)/sizeof(msemaphore))+1)

/* Barrier structure */
typedef struct parmacs_barrier {
  msemaphore   s;          /* msemaphore for mutual exclusion */
  union {
    msemaphore pad[MSEM_PAD(3*sizeof(int))];
    struct {
      volatile int n;      /* number of processes waiting */
      volatile int seq;    /* sequence number for barrier operation */
      volatile int active; /* barrier is being used */
    } data;
  } u;
} parmacs_barrier;

/* Global subscript structure */
typedef struct parmacs_global {
  msemaphore   s; /* semaphore for mutual exclusion */
  union {
    msemaphore pad[MSEM_PAD(4*sizeof(int))];
    struct {
      volatile int i;      /* next subscript */
      volatile int n;      /* number of processes finished */
      volatile int seq;    /* sequence number for global subscript operations */
      volatile int active; /* global subscript is being used */
    } data;
  } u;
} parmacs_global;

/* Pause/event structure */
typedef struct parmacs_event {
  msemaphore s;            /* semaphore for mutual exclusion */
  union {
    msemaphore pad[MSEM_PAD(sizeof(int) + sizeof(void *))];
    struct {
      volatile int v;      /* event value */
      void *blocked;       /* processes waiting for event to be set */
    } data;
  } u;
} parmacs_event;

/* lock structure */
typedef struct parmacs_lock {
  msemaphore s;            /* Semaphore */
} parmacs_lock;

/* Time stamps */
typedef unsigned long int parmacs_timestamp;

/* Processes */
typedef unsigned int parmacs_pid;

#ifdef PARMACS_INTERNALS
typedef struct parmacs_process {
  struct parmacs_process *link;  /* Queues */
  volatile parmacs_event *event; /* waiting for this event */
  int          barrier_seq;      /* sequence number of current barrier op. */
  int          global_seq;       /* sequence number of current glob. subs. op.*/
} parmacs_process;

typedef parmacs_process *parmacs_process_t;

/* Parmacs control structure */
#define PARMACS_MAIN_ID    0
#define PARMACS_CONFIG_ID  1
#define PARMACS_MEMORY_ID  2
#define PARMACS_BARRIER_ID 3
#define PARMACS_GLOBAL_ID  4
#define PARMACS_EVENT_ID   5
#define PARMACS_LOCK_ID    6
#define PARMACS_PROCESS_ID 7
#define PARMACS_MAX_ID     8
typedef struct parmacs_data {
  parmacs_timestamp start_time;  /* Starting time */
  parmacs_timestamp end_time;    /* Ending time */
#ifdef PARMACS_NO_SCHED_YIELD
  int               blocking_fd; /* Blocking fd to implement processor yield */
#endif /* PARMACS_NO_SCHED_YIELD */
  int               mmap_fd;     /* File descriptor for mmaps */ 
  parmacs_barrier  *barrier;     /* Barrier structures */
  int               n_barriers;
  parmacs_global   *global;      /* Global subscript structures */
  int               n_globals;
  parmacs_event    *event;       /* Event structures */
  int               n_events;
  parmacs_lock     *lock;        /* Lock structures */
  int               n_locks;
  parmacs_process  *process;     /* Parmacs processes */
  parmacs_process_t free_process;/* Free processes queue */
  msemaphore       *mutex;       /* Semaphores for mutual exclusions */
  void             *shm_address; /* Beginning of the shared memory segment */
  unsigned long     shm_size;    /* Size of the shared memory area */
  unsigned long     shm_avail;   /* Amount of available shared memory */
  void             *shm_pointer; /* Pointer to free memory */
} parmacs_data;

typedef parmacs_data *parmacs_data_t;

#endif /* PARMACS_INTERNALS */

#endif /* _BSD_PARMACS_TYPES_H_ */
