/*
 * File: lwp/parmacs_support.c
 * Description: implementation of support code for lwp based PARMACS
 *   macros. Also define global var definitions.
 * Comments: 
 * Copyright: Department of Computer Architecture - Polytechnic University of
 *   Catalonia (UPC)
 * Created: 1999/06 ernest@ac.upc.es (E.Artiaga)
 * Contact: gsomk@ac.upc.es
 * Last update:
 * History:
 */
/*
 * Disclaimer and copying policy:
 *
 * The Polytechnic University of Catalonia (UPC) and the authors disclaim
 * all warranties with regard to this software, including all implied
 * warranties of merchantability and fitness.
 * In no event shall the Polytechnic University of Catalonia (UPC) or the
 * authors be liable for any special, indirect or consequential damages or
 * any damages whatsoever resulting from loss of use, data or profits,
 * whether in an action of contract, negligence or other tortius action,
 * arising out of or in connection with the use or performance of this
 * software.
 *
 * Permission to use, copy, modify, and distribute this software for
 * any purpose and without fee is hereby granted, provided that the above
 * copyright notice and disclaimer, and also this permission notice, appear
 * in all copies.
 *
 */

#include <parmacs_internals.h>

/*
 * Data structures
 */
 
parmacs_data_t _parmacs_data = (parmacs_data_t)0;
parmacs_configuration _parmacs_configuration = {
  (char *)0,
  PARMACS_SPIN_COUNT,
  PARMACS_MAX_BARRIER,
  PARMACS_MAX_GLOBAL,
  PARMACS_MAX_EVENT,
  PARMACS_MAX_LOCK,
  PARMACS_PAGE_SIZE,
  PARMACS_STACK_SIZE,
  PARMACS_SHARED_MEMORY,
  PARMACS_MEM_ALIGNMENT,
  PARMACS_MAX_PROCESS,
  PARMACS_TIMESTAMP
};
parmacs_configuration_t _parmacs_config = &_parmacs_configuration;

static void _parmacs_configure(unsigned long int);
static void _parmacs_process_wrapper(void *);

/*
 * Initialize environment
 * Comments:
 *   This routine is supposed to be called from the main routine, at the
 *   very beginning of the execution.
 */
void
_parmacs_main_initenv(unsigned long int shm_size)
{
  register void *address;
  register int i;
  register parmacs_process_t myself;
  
  /* Load configuration */
  _parmacs_configure(shm_size);
     
  /* Allocate memory for parmacs data */
  /* MAIN DATA */
  address = malloc(sizeof(parmacs_data));
  if (!address) {
    _parmacs_panic("parmacs: cannot allocate memory for parmacs data!");
  }
  _parmacs_data = (parmacs_data_t)address;
  
  /* Initialize mutex locks */
  address = malloc(sizeof(lwp_mutex_t)*PARMACS_MAX_ID);
  if (!address) {
    _parmacs_panic("parmacs: cannot allocate semaphores for mutual exclusion!");
  }
  _parmacs_data->mutex = address;
  for (i = 0; i < PARMACS_MAX_ID; i++) {
    _lwp_mutex_init(&(_parmacs_data->mutex[i]));
  }
    
  /* BARRIERS */
  address = malloc(_parmacs_config->max_barriers*sizeof(parmacs_barrier));
  if (!address) {
    _parmacs_panic("parmacs: cannot allocate memory for barriers!");
  }
  _parmacs_data->barrier = address;
  _parmacs_data->n_barriers = 0;
  
  /* GLOBAL SUBSCRIPTS */
  address = malloc(_parmacs_config->max_globals*sizeof(parmacs_global));
  if (!address) {
    _parmacs_panic("parmacs: cannot allocate memory for global subscripts!");
  }
  _parmacs_data->global = address;
  _parmacs_data->n_globals = 0;
  
  /* EVENTS */
  address = malloc(_parmacs_config->max_events*sizeof(parmacs_event));
  if (!address) {
    _parmacs_panic("parmacs: cannot allocate memory for events!");
  }
  _parmacs_data->event = address;
  _parmacs_data->n_events = 0;
  
  /* LOCKS */
  address = malloc(_parmacs_config->max_locks*sizeof(parmacs_lock));
  if (!address) {
    _parmacs_panic("parmacs: cannot allocate memory for locks!");
  }
  _parmacs_data->lock = address;
  _parmacs_data->n_locks = 0;
  
  /* PROCESSES */
  address = malloc(_parmacs_config->max_processes*sizeof(parmacs_process));
  if (!address) {
    _parmacs_panic("parmacs: cannot allocate memory for processes!");
  }
  _parmacs_data->process = address;
  for(i=1; i < (_parmacs_config->max_processes-1); i++) {
    _parmacs_data->process[i].link = &(_parmacs_data->process[i+1]);
    _parmacs_data->process[i].ref_count = 0;
  }
  _parmacs_data->process[_parmacs_config->max_processes-1].link =
      (parmacs_process_t)0;
  _parmacs_data->process[_parmacs_config->max_processes-1].ref_count = 0;
  _parmacs_data->free_process = &(_parmacs_data->process[1]);
  myself = &(_parmacs_data->process[0]);
  myself->ref_count = 1;
  myself->children  = 0;
  myself->finished  = 0;
  myself->parent    = (parmacs_process_t)0;
  _lwp_mutex_init(&(myself->m));
  _lwp_cond_init(&(myself->c));
  _lwp_setprivate((void *)myself);
  
  /*
   * Time stamp:
   */
  if (_parmacs_config->timestamp) {
    struct tms now;
    _parmacs_data->start_time = times(&now);
  }
}

void
_parmacs_main_end()
{
  /*
   * Time stamp:
   */
  if (_parmacs_config->timestamp) {
    struct tms now;
    _parmacs_data->end_time = times(&now);
    fprintf(stderr, "\nPARMACS TIME: start: %lu, end: %lu, elapsed: %lu.\n",
            _parmacs_data->start_time,
            _parmacs_data->end_time,
            _parmacs_data->end_time - _parmacs_data->start_time);
  }
  _parmacs_exit();
}

static void
_parmacs_configure(unsigned long int shm_size)
{
  extern char **environ;
  char **my_environment;
  char * var_name;

  _parmacs_configuration.shared_memory = shm_size;
  
  /* Read configuration from environment */
  
  for(my_environment = environ; *my_environment; my_environment++) {
    var_name = strtok(strdup(*my_environment), "=");
    if (!strcmp(var_name, "PARMACS_CONFIG_FILE")) {
      _parmacs_configuration.config_file = strdup(strtok(NULL, "\0"));
    } else if (!strcmp(var_name, "PARMACS_SPIN_COUNT")) {
      _parmacs_configuration.spin_count = atoi(strtok(NULL, "\0"));
    } else if (!strcmp(var_name, "PARMACS_MAX_BARRIER")) {
      _parmacs_configuration.max_barriers = atoi(strtok(NULL, "\0"));
    } else if (!strcmp(var_name, "PARMACS_MAX_GLOBAL")) {
      _parmacs_configuration.max_globals = atoi(strtok(NULL, "\0"));
    } else if (!strcmp(var_name, "PARMACS_MAX_EVENT")) {
      _parmacs_configuration.max_events = atoi(strtok(NULL, "\0"));
    } else if (!strcmp(var_name, "PARMACS_MAX_LOCK")) {
      _parmacs_configuration.max_locks = atoi(strtok(NULL, "\0"));
    } else if (!strcmp(var_name, "PARMACS_PAGE_SIZE")) {
      _parmacs_configuration.page_size = atoi(strtok(NULL, "\0"));
    } else if (!strcmp(var_name, "PARMACS_STACK_SIZE")) {
      _parmacs_configuration.stack_size = atoi(strtok(NULL, "\0"));
    } else if (!strcmp(var_name, "PARMACS_SHARED_MEMORY")) {
      char *endptr;
      _parmacs_configuration.shared_memory = 
        strtoul(strtok(NULL, "\0"), &endptr, 10);
    } else if (!strcmp(var_name, "PARMACS_MEM_ALIGNMENT")) {
      _parmacs_configuration.memory_alignment = atoi(strtok(NULL, "\0"));
    } else if (!strcmp(var_name, "PARMACS_MAX_PROCESS")) {
      _parmacs_configuration.max_processes = atoi(strtok(NULL, "\0"));
    } else if (!strcmp(var_name, "PARMACS_TIMESTAMP")) {
      _parmacs_configuration.timestamp = atoi(strtok(NULL, "\0"));
    }
    free(var_name);  
  }
  
  /* if there is a configuration file, use it */
  if (_parmacs_configuration.config_file) {
    char buffer[512];
    FILE *fd;

    if (!(fd = fopen(_parmacs_configuration.config_file, "r"))) {
      _parmacs_panic("parmacs: cannot open configuration file!");
    }    
    while (fgets(buffer, 512, fd)) {
      var_name = strtok(buffer, " =\t\n\0");
      if (!strcmp(var_name, "PARMACS_SPIN_COUNT")) {
        _parmacs_configuration.spin_count = atoi(strtok(NULL, " =\t\n\0"));
      } else if (!strcmp(var_name, "PARMACS_MAX_BARRIER")) {
        _parmacs_configuration.max_barriers = atoi(strtok(NULL, " =\t\n\0"));
      } else if (!strcmp(var_name, "PARMACS_MAX_GLOBAL")) {
        _parmacs_configuration.max_globals = atoi(strtok(NULL, " =\t\n\0"));
      } else if (!strcmp(var_name, "PARMACS_MAX_EVENT")) {
        _parmacs_configuration.max_events = atoi(strtok(NULL, " =\t\n\0"));
      } else if (!strcmp(var_name, "PARMACS_MAX_LOCK")) {
        _parmacs_configuration.max_locks = atoi(strtok(NULL, " =\t\n\0"));
      } else if (!strcmp(var_name, "PARMACS_PAGE_SIZE")) {
        _parmacs_configuration.page_size = atoi(strtok(NULL, " =\t\n\0"));
      } else if (!strcmp(var_name, "PARMACS_STACK_SIZE")) {
        _parmacs_configuration.stack_size = atoi(strtok(NULL, " =\t\n\0"));
      } else if (!strcmp(var_name, "PARMACS_SHARED_MEMORY")) {
        char *endptr;
        _parmacs_configuration.shared_memory = 
          strtoul(strtok(NULL, " =\t\n\0"), &endptr, 10);
      } else if (!strcmp(var_name, "PARMACS_MEM_ALIGNMENT")) {
        _parmacs_configuration.memory_alignment=atoi(strtok(NULL, " =\t\n\0"));
      } else if (!strcmp(var_name, "PARMACS_MAX_PROCESS")) {
        _parmacs_configuration.max_processes = atoi(strtok(NULL, "\0"));
      } else if (!strcmp(var_name, "PARMACS_TIMESTAMP")) {
        _parmacs_configuration.timestamp = atoi(strtok(NULL, " =\t\n\0"));
      }
    }
    fclose(fd);
  }
}

/*
 * Memory management
 * Comments:
 *   Memory is shared among pthreads. So, simple malloc is enough
 */
 
void *
_parmacs_memory_alloc(int size)
{
  register void *address;
  register unsigned long int u_ptr;
  register int adjust;
  
  /* malloc is not reentrant */
  _parmacs_internal_lock(PARMACS_MEMORY_ID);
  address = (void *)malloc(size+_parmacs_config->memory_alignment);
  _parmacs_internal_unlock(PARMACS_MEMORY_ID);
  u_ptr = (unsigned long int)address;
  if ((adjust = u_ptr & (_parmacs_config->memory_alignment-1))) {
    u_ptr += _parmacs_config->memory_alignment - adjust;
    address = (void *)u_ptr;
  }
  return (void *)address;
}

void
_parmacs_memory_free(void *address)
{
  /* not implemented */
}

void *
_parmacs_local_alloc(int size)
{
  register void *address;
  /* malloc is not reentrant: lock needed */
  _parmacs_internal_lock(PARMACS_MEMORY_ID);
  address = (void *)malloc(size);
  _parmacs_internal_unlock(PARMACS_MEMORY_ID);
  return address;
}

void
_parmacs_local_free(void *address)
{
  free(address);
}

/*
 * Process creation and joining
 */
 
static void
_parmacs_process_wrapper(void *pr_data)
{
  register parmacs_process_t process = (parmacs_process_t)pr_data;
  register parmacs_process_t parent;
  
  /* self-identification */
  _lwp_setprivate(pr_data);
  parent = process->parent;
  
  /* execute code */
  process->routine();
  
  /* Notify parent and clean up */
  _parmacs_internal_lock(PARMACS_PROCESS_ID);
  _lwp_mutex_lock(&(parent->m));
  parent->finished ++;
  parent->ref_count --;
  if (!parent->ref_count) {
    parent->link = _parmacs_data->free_process;
    _parmacs_data->free_process = parent;
  }
  _lwp_cond_signal(&(parent->c));
  _lwp_mutex_unlock(&(parent->m));
  _lwp_mutex_lock(&(process->m));
  process->ref_count --;
  if (!process->ref_count) {
    process->link = _parmacs_data->free_process;
    _parmacs_data->free_process = process;
  }
  _lwp_mutex_unlock(&(process->m));
  _parmacs_internal_unlock(PARMACS_PROCESS_ID);
  _lwp_exit();
}
 
void
_parmacs_create(void (*routine)())
{
  register int i;
  parmacs_process_t myself, child;
  lwpid_t new;
  ucontext_t *contextp;
  void * stack;

  /* self identify */
  myself = (parmacs_process_t)_lwp_getprivate();

  /* Allocate an entry for the new process */
  _parmacs_internal_lock(PARMACS_PROCESS_ID);
  if (!_parmacs_data->free_process) {
    _parmacs_internal_unlock(PARMACS_PROCESS_ID);
    _parmacs_panic("parmacs: too many processes!\n");
  }
  child = _parmacs_data->free_process;
  _parmacs_data->free_process = child->link;
  child->ref_count = 1;
  _parmacs_internal_unlock(PARMACS_PROCESS_ID);

  /* Initialize */
  _lwp_mutex_init(&(child->m));
  _lwp_cond_init(&(child->c));
  child->children = 0;
  child->finished = 0;
  child->parent   = myself;
  child->routine  = routine;
  child->link     = (parmacs_process_t)0;
  
  /* increase parent's reference count */
  _lwp_mutex_lock(&(myself->m));
  myself->ref_count ++;
  _lwp_mutex_unlock(&(myself->m));
  
  /* Actually fork */
  contextp = (ucontext_t *)_parmacs_local_alloc(sizeof(ucontext_t));
  stack = (void *)_parmacs_local_alloc(
      _parmacs_config->stack_size * _parmacs_config->page_size);
  sigprocmask(SIG_SETMASK, NULL, &contextp->uc_sigmask);
  _lwp_makecontext(contextp, _parmacs_process_wrapper,
      (void *)child, (void *)child,
      stack, _parmacs_config->stack_size *_parmacs_config->page_size);
  if (_lwp_create(contextp, LWP_DETACHED, &new)) {
    _parmacs_panic("parmacs: cannot create new process!");
  }
}

void
_parmacs_wait_for_end(int amount)
{
  parmacs_process_t myself;
  
  myself = _lwp_getprivate();
  _lwp_mutex_lock(&(myself->m));
  while(amount) {
    if (myself->finished) {
      amount --;
      myself->finished --;
      myself->children --;
    } else {
      _lwp_cond_wait(&(myself->c), &(myself->m));
    } 
  }
  _lwp_mutex_unlock(&(myself->m));
}

parmacs_pid
_parmacs_get_pid()
{
  return ((parmacs_pid)_lwp_self());
}

/*
 * Scheduling, clean up and exit routines
 */
 
void 
_parmacs_panic(char *msg)
{
  if (msg) fprintf(stderr, "%s\n", msg);
  exit(-1);
}

void 
_parmacs_exit()
{
  exit(0);
}

void 
_parmacs_yield()
{
  sched_yield();
}

/*
 * Clock
 */
parmacs_timestamp
_parmacs_clock()
{
  struct tms now;
  return (parmacs_timestamp)times(&now);
}

