/*
 * File: lwp/parmacs_sync.c
 * Description: implementation of synchronization macros of PARMACS
 * Comments: 
 * Copyright: Department of Computer Architecture - Polytechnic University of
 *   Catalonia (UPC)
 * Created: 1999/06 ernest@ac.upc.es (E.Artiaga)
 * Contact: gsomk@ac.upc.es
 * Last update:
 * History:
 */
/*
 * Disclaimer and copying policy:
 *
 * The Polytechnic University of Catalonia (UPC) and the authors disclaim
 * all warranties with regard to this software, including all implied
 * warranties of merchantability and fitness.
 * In no event shall the Polytechnic University of Catalonia (UPC) or the
 * authors be liable for any special, indirect or consequential damages or
 * any damages whatsoever resulting from loss of use, data or profits,
 * whether in an action of contract, negligence or other tortius action,
 * arising out of or in connection with the use or performance of this
 * software.
 *
 * Permission to use, copy, modify, and distribute this software for
 * any purpose and without fee is hereby granted, provided that the above
 * copyright notice and disclaimer, and also this permission notice, appear
 * in all copies.
 *
 */

#include <parmacs_internals.h>


/*
 * Barriers
 */

void
_parmacs_barinit(parmacs_barrier_t *barrier)
{
  /* get a free barrier entry */
  _parmacs_internal_lock(PARMACS_BARRIER_ID);
  if (_parmacs_data->n_barriers >= _parmacs_config->max_barriers) {
    _parmacs_internal_unlock(PARMACS_BARRIER_ID);
    _parmacs_panic("parmacs: too many barriers!");
  }
  *barrier = &_parmacs_data->barrier[_parmacs_data->n_barriers++];
  _parmacs_internal_unlock(PARMACS_BARRIER_ID);
  
  /* now, initialize the new barrier */
  _lwp_mutex_init(&((*barrier)->m));
  _lwp_cond_init(&((*barrier)->c));
  (*barrier)->n = 0;
}
 
void 
_parmacs_barrier(parmacs_barrier_t barrier, int procs)
{
  _lwp_mutex_lock(&(barrier->m));
  if (++(barrier->n) < procs) {
    _lwp_cond_wait(&(barrier->c), &(barrier->m));
  } else {
    barrier->n = 0;
    _lwp_cond_broadcast(&(barrier->c));
  }
  _lwp_mutex_unlock(&(barrier->m));
}

/*
 * Global subscripts
 */
 
void
_parmacs_gsinit(parmacs_global_t *global)
{
  /* get a free global subscript entry */
  _parmacs_internal_lock(PARMACS_GLOBAL_ID);
  if (_parmacs_data->n_globals >= _parmacs_config->max_globals) {
    _parmacs_internal_unlock(PARMACS_GLOBAL_ID);
    _parmacs_panic("parmacs: too many global subscripts!");
  }
  *global = &_parmacs_data->global[_parmacs_data->n_globals++];
  _parmacs_internal_unlock(PARMACS_GLOBAL_ID);
  
  /* now, initialize the new global subscript */
  _lwp_mutex_init(&((*global)->m));
  _lwp_cond_init(&((*global)->c));
  (*global)->i = 0;
  (*global)->n = 0;
}

void
_parmacs_getsub(parmacs_global_t global, int *next, int last, int procs)
{
  _lwp_mutex_lock(&(global->m));
  if (global->i <= last) {
    *next = (global->i)++;
  } else {
    *next = -1;
    if (++(global->n) == procs) {
      global->i = 0;
      global->n = 0;
      _lwp_cond_broadcast(&(global->c));
    } else {
      _lwp_cond_wait(&(global->c), &(global->m));
    }
  }
  _lwp_mutex_unlock(&(global->m));
}

/*
 * Locks
 */

void 
_parmacs_lockinit(parmacs_lock_t *lock)
{
  /* get a free lock entry */
  _parmacs_internal_lock(PARMACS_LOCK_ID);
  if (_parmacs_data->n_locks >= _parmacs_config->max_locks) {
    _parmacs_internal_unlock(PARMACS_LOCK_ID);
    _parmacs_panic("parmacs: too many locks!");
  }
  *lock = &_parmacs_data->lock[_parmacs_data->n_locks++];
  _parmacs_internal_unlock(PARMACS_LOCK_ID);
  
  /* now, initialize the new lock */
  _lwp_mutex_init(&((*lock)->m));
}

void
_parmacs_lock(parmacs_lock_t lock)
{
  _parmacs_sem_lock(&(lock->m));
}

void
_parmacs_unlock(parmacs_lock_t lock)
{
  _parmacs_sem_unlock(&(lock->m));
}

void
_parmacs_alockinit(parmacs_lock_t *lock, int amount)
{
  register int i;
  
  /* get a free lock entry */
  _parmacs_internal_lock(PARMACS_LOCK_ID);
  if ((_parmacs_data->n_locks + amount -1) >= _parmacs_config->max_locks) {
    _parmacs_internal_unlock(PARMACS_LOCK_ID);
    _parmacs_panic("parmacs: too many locks!");
  }
  *lock = &_parmacs_data->lock[_parmacs_data->n_locks];
  _parmacs_data->n_locks += amount;
  _parmacs_internal_unlock(PARMACS_LOCK_ID);
  
  /* now, initialize the new locks */
  for (i=0; i < amount; i++) {
    _lwp_mutex_init(&((*lock)[i].m));
  }
}

void
_parmacs_alock(parmacs_lock_t lock, int idx)
{
  _parmacs_sem_lock(&(lock[idx].m));
}

void
_parmacs_aulock(parmacs_lock_t lock, int idx)
{
  _parmacs_sem_unlock(&(lock[idx].m));
}

/*
 * Events
 */
 
void
_parmacs_pauseinit(parmacs_event_t *event, int amount)
{
  register int i;
  
  /* get the free event entries */
  _parmacs_internal_lock(PARMACS_EVENT_ID);
  if ((_parmacs_data->n_events + amount -1) >= _parmacs_config->max_events) {
    _parmacs_internal_unlock(PARMACS_EVENT_ID);
    _parmacs_panic("parmacs: too many events!");
  }
  *event = &_parmacs_data->event[_parmacs_data->n_events];
  _parmacs_data->n_events += amount;
  _parmacs_internal_unlock(PARMACS_EVENT_ID);
  
  /* initialize the event entries */
  for(i=0; i < amount; i++) {
    _lwp_mutex_init(&((*event)[i].m));
    _lwp_cond_init(&((*event)[i].c));
    _lwp_cond_init(&((*event)[i].s));
    (*event)[i].v = 0;
  }
}

void
_parmacs_clearpause(parmacs_event_t event, int idx)
{
  _lwp_mutex_lock(&(event[idx].m));
  event[idx].v = 0;
  _lwp_cond_broadcast(&(event[idx].c));
  _lwp_mutex_unlock(&(event[idx].m));
}

void
_parmacs_setpause(parmacs_event_t event, int idx)
{
  _lwp_mutex_lock(&(event[idx].m));
  event[idx].v = 1;
  _lwp_cond_broadcast(&(event[idx].s));
  _lwp_mutex_unlock(&(event[idx].m));
}

void
_parmacs_event(parmacs_event_t event, int idx)
{
  _lwp_mutex_lock(&(event[idx].m));
  while (event[idx].v == 1) {
    _lwp_cond_wait(&(event[idx].c), &(event[idx].m));
  }
  event[idx].v = 1;
  _lwp_cond_broadcast(&(event[idx].s));
  _lwp_mutex_unlock(&(event[idx].m));
}


void
_parmacs_waitpause(parmacs_event_t event, int idx)
{
  _lwp_mutex_lock(&(event[idx].m));
  if (event[idx].v == 0) {
    _lwp_cond_wait(&(event[idx].s), &(event[idx].m));
  }
  _lwp_mutex_unlock(&(event[idx].m));
}

void
_parmacs_pause(parmacs_event_t event, int idx)
{
  _lwp_mutex_lock(&(event[idx].m));
  while (event[idx].v == 0) {
    _lwp_cond_wait(&(event[idx].s), &(event[idx].m));
  }
  event[idx].v = 0;
  _lwp_cond_broadcast(&(event[idx].c));
  _lwp_mutex_unlock(&(event[idx].m));
}  
  
  
