/*
 * File: pthreads.sparc/parmacs_sync.c
 * Description: implementation of synchronization macros of PARMACS
 * Comments: 
 * Copyright: Department of Computer Architecture - Polytechnic University of
 *   Catalonia (UPC)
 * Created: 1999/06 ernest@ac.upc.es (E.Artiaga)
 * Contact: gsomk@ac.upc.es
 * Last update:
 * History:
 */
/*
 * Disclaimer and copying policy:
 *
 * The Polytechnic University of Catalonia (UPC) and the authors disclaim
 * all warranties with regard to this software, including all implied
 * warranties of merchantability and fitness.
 * In no event shall the Polytechnic University of Catalonia (UPC) or the
 * authors be liable for any special, indirect or consequential damages or
 * any damages whatsoever resulting from loss of use, data or profits,
 * whether in an action of contract, negligence or other tortius action,
 * arising out of or in connection with the use or performance of this
 * software.
 *
 * Permission to use, copy, modify, and distribute this software for
 * any purpose and without fee is hereby granted, provided that the above
 * copyright notice and disclaimer, and also this permission notice, appear
 * in all copies.
 *
 */

#include <parmacs_internals.h>


/*
 * Barriers
 */

void
_parmacs_barinit(parmacs_barrier_t *barrier)
{
  /* get a free barrier entry */
  _parmacs_internal_lock(PARMACS_BARRIER_ID);
  if (_parmacs_data->n_barriers >= _parmacs_config->max_barriers) {
    _parmacs_internal_unlock(PARMACS_BARRIER_ID);
    _parmacs_panic("parmacs: too many barriers!");
  }
  *barrier = &_parmacs_data->barrier[_parmacs_data->n_barriers++];
  _parmacs_internal_unlock(PARMACS_BARRIER_ID);
  
  /* now, initialize the new barrier */
  (*barrier)->s = 0;
  (*barrier)->n = 0;
  (*barrier)->seq = 0;
  (*barrier)->active = 0;
}
 
void 
_parmacs_barrier(parmacs_barrier_t barrier, int procs)
{
  register int countdown;
  register int spin = _parmacs_config->busy_wait_count;
  parmacs_process_t myself;
  
  /* self identify */
  _posix_getspecific(_parmacs_data->private, (void **)&myself);

  /* Enter the barrier */
  _parmacs_sem_lock(&(barrier->s));
  if (!barrier->active) {
    /* Setup the barrier */
    barrier->active = 1;
    barrier->seq++;
  }
  
  barrier->n++;
  myself->barrier_seq = barrier->seq;
  
  if (barrier->n == procs) {
    /* This is the last process - reset the barrier and wakeup everybody */
    barrier->n = 0;
    barrier->active = 0;
  }
  _parmacs_sem_unlock(&(barrier->s));
  
  /* now, wait for everybody to finish, and then go on */
  countdown = spin;
  while (barrier->active && 
         barrier->seq == myself->barrier_seq) {
    if (countdown >= 0) {
      if (!countdown) {
        _parmacs_yield();
        countdown = spin;
      } else {
        countdown --;
      }
    }
  }
}
  
/*
 * Global subscripts
 */
 
void
_parmacs_gsinit(parmacs_global_t *global)
{
  /* get a free global subscript entry */
  _parmacs_internal_lock(PARMACS_GLOBAL_ID);
  if (_parmacs_data->n_globals >= _parmacs_config->max_globals) {
    _parmacs_internal_unlock(PARMACS_GLOBAL_ID);
    _parmacs_panic("parmacs: too many global subscripts!");
  }
  *global = &_parmacs_data->global[_parmacs_data->n_globals++];
  _parmacs_internal_unlock(PARMACS_GLOBAL_ID);
  
  /* now, initialize the new global subscript */
  (*global)->s = 0;
  (*global)->i = 0;
  (*global)->n = 0;
  (*global)->seq = 0;
  (*global)->active = 0;
}

void
_parmacs_getsub(parmacs_global_t global, int *next, int last, int procs)
{
  register int countdown;
  register int spin = _parmacs_config->busy_wait_count;
  parmacs_process_t myself;
  
  /* self identify */
  _posix_getspecific(_parmacs_data->private, (void **)&myself);
    
  _parmacs_sem_lock(&(global->s));
  
  if (!global->active) {
    /* First to enter: setup global subscript */
    global->active = 1;
    global->seq ++;
  }

  /* Get next subscript */
  if (global->i <= last) {
    *next = global->i++;
    _parmacs_sem_unlock(&(global->s));
  } else {
    *next = -1;
    /* the loop is finished: wait until everybody is done */
    global->n++;
    myself->global_seq = global->seq;
    
    if (global->n == procs) {
      /* This is the last process - reset subscript and wakeup everybody */
      global->i = 0;
      global->n = 0;
      global->active = 0;
    }
    _parmacs_sem_unlock(&(global->s));

    /* now, wait for everybody to finish, and then go on */
    countdown = spin;
    while (global->active && 
           global->seq == myself->global_seq) {
      if (countdown >= 0) {
        if (!countdown) {
          _parmacs_yield();
          countdown = spin;
        } else {
          countdown --;
        }
      }
    }
  }
}

/*
 * Locks
 */

void 
_parmacs_lockinit(parmacs_lock_t *lock)
{
  /* get a free lock entry */
  _parmacs_internal_lock(PARMACS_LOCK_ID);
  if (_parmacs_data->n_locks >= _parmacs_config->max_locks) {
    _parmacs_internal_unlock(PARMACS_LOCK_ID);
    _parmacs_panic("parmacs: too many locks!");
  }
  *lock = &_parmacs_data->lock[_parmacs_data->n_locks++];
  _parmacs_internal_unlock(PARMACS_LOCK_ID);
  
  /* now, initialize the new lock */
  (*lock)->s = 0;
}

void
_parmacs_lock(parmacs_lock_t lock)
{
  _parmacs_sem_lock(&(lock->s));
}

void
_parmacs_unlock(parmacs_lock_t lock)
{
  _parmacs_sem_unlock(&(lock->s));
}

void
_parmacs_alockinit(parmacs_lock_t *lock, int amount)
{
  register int i;
  
  /* get a free lock entry */
  _parmacs_internal_lock(PARMACS_LOCK_ID);
  if ((_parmacs_data->n_locks + amount -1) >= _parmacs_config->max_locks) {
    _parmacs_internal_unlock(PARMACS_LOCK_ID);
    _parmacs_panic("parmacs: too many locks!");
  }
  *lock = &_parmacs_data->lock[_parmacs_data->n_locks];
  _parmacs_data->n_locks += amount;
  _parmacs_internal_unlock(PARMACS_LOCK_ID);
  
  /* now, initialize the new locks */
  for (i=0; i < amount; i++) {
    (*lock)[i].s = 0;
  }
}

void
_parmacs_alock(parmacs_lock_t lock, int idx)
{
  _parmacs_sem_lock(&(lock[idx].s));
}

void
_parmacs_aulock(parmacs_lock_t lock, int idx)
{
  _parmacs_sem_unlock(&(lock[idx].s));
}

/*
 * Events
 */
 
void
_parmacs_pauseinit(parmacs_event_t *event, int amount)
{
  register int i;
  
  /* get the free event entries */
  _parmacs_internal_lock(PARMACS_EVENT_ID);
  if ((_parmacs_data->n_events + amount -1) >= _parmacs_config->max_events) {
    _parmacs_internal_unlock(PARMACS_EVENT_ID);
    _parmacs_panic("parmacs: too many events!");
  }
  *event = &_parmacs_data->event[_parmacs_data->n_events];
  _parmacs_data->n_events += amount;
  _parmacs_internal_unlock(PARMACS_EVENT_ID);
  
  /* initialize the event entries */
  for(i=0; i < amount; i++) {
    (*event)[i].s = 0;
    (*event)[i].v = 0;
    (*event)[i].blocked = (void *)0;
  }
}

void
_parmacs_clearpause(parmacs_event_t event, int idx)
{
  /* must be locked to avoid interferences with other event operations */
  _parmacs_sem_lock(&(event[idx].s));
  event[idx].v = 0;
  _parmacs_sem_unlock(&(event[idx].s));
}

void
_parmacs_setpause(parmacs_event_t event, int idx)
{
  register int i;
  register parmacs_process_t process;
   
  _parmacs_sem_lock(&(event[idx].s));
  /* set the value */
  event[idx].v = 1;
  /* wakeup everybody waiting for this event */
  process = (parmacs_process_t)(event[idx].blocked);
  while (process) {
    event[idx].blocked = (void *)(process->link);
    process->link  = (parmacs_process_t)0;
    process->event = (parmacs_event_t)0;
    process = (parmacs_process_t)(event[idx].blocked);
  }
  _parmacs_sem_unlock(&(event[idx].s));
}

void
_parmacs_event(parmacs_event_t event, int idx)
{
  register int countdown;
  register int spin = _parmacs_config->busy_wait_count;
  register parmacs_process_t process;

  /* Wait for the event to be zero */  
  _parmacs_sem_lock(&(event[idx].s));
  while(event[idx].v == 1) {
    _parmacs_sem_unlock(&(event[idx].s));
    countdown = spin;
    while(event[idx].v == 1) {
      if (countdown >= 0) {
        if (!countdown) {
          _parmacs_yield();
          countdown = spin;
        } else {
          countdown --;
        }
      }
    }
    _parmacs_sem_lock(&(event[idx].s));
  }
  
  /* set the event again */
  event[idx].v = 1;    

  /* wakeup everybody waiting for this event to be set */
  process = (parmacs_process_t)(event[idx].blocked);
  while (process) {
    event[idx].blocked = (void *)(process->link);
    process->link  = (parmacs_process_t)0;
    process->event = (parmacs_event_t)0;
    process = (parmacs_process_t)(event[idx].blocked);
  }
  _parmacs_sem_unlock(&(event[idx].s));
}


void
_parmacs_waitpause(parmacs_event_t event, int idx)
{
  register int countdown;
  register int spin = _parmacs_config->busy_wait_count;
  parmacs_process_t myself;
  
  /* self identify */
  _posix_getspecific(_parmacs_data->private, (void **)&myself);
  
  /* check the event and prepare to block */
  _parmacs_sem_lock(&(event[idx].s));
  if (event[idx].v == 0) {
    myself->link = (parmacs_process_t)(event[idx].blocked);
    event[idx].blocked = (void *)myself;
    myself->event = &event[idx];
  }
  _parmacs_sem_unlock(&(event[idx].s));
  
  /* now, wait */
  countdown = spin;
  while(myself->event) {
    if (countdown >= 0) {
      if (!countdown) {
        _parmacs_yield();
        countdown = spin;
      } else {
        countdown --;
      }
    }
  }
}

void
_parmacs_pause(parmacs_event_t event, int idx)
{
  register int countdown;
  register int spin = _parmacs_config->busy_wait_count;

  /* Wait for the event to be set */  
  _parmacs_sem_lock(&(event[idx].s));
  while(event[idx].v == 0) {
    _parmacs_sem_unlock(&(event[idx].s));
    countdown = spin;
    while(event[idx].v == 0) {
      if (countdown >= 0) {
        if (!countdown) {
          _parmacs_yield();
          countdown = spin;
        } else {
          countdown --;
        }
      }
    }
    _parmacs_sem_lock(&(event[idx].s));
  }
  
  /* zero the event again */
  event[idx].v = 0;    
  _parmacs_sem_unlock(&(event[idx].s));
}  
  
  
