/*
 * File: sysv/parmacs_support.c
 * Description: implementation of support code for System V based PARMACS
 *   macros. Also define global var definitions.
 * Comments: 
 * Copyright: Department of Computer Architecture - Polytechnic University of
 *   Catalonia (UPC)
 * Created: 1999/06 ernest@ac.upc.es (E.Artiaga)
 * Contact: gsomk@ac.upc.es
 * Last update:
 * History:
 */
/*
 * Disclaimer and copying policy:
 *
 * The Polytechnic University of Catalonia (UPC) and the authors disclaim
 * all warranties with regard to this software, including all implied
 * warranties of merchantability and fitness.
 * In no event shall the Polytechnic University of Catalonia (UPC) or the
 * authors be liable for any special, indirect or consequential damages or
 * any damages whatsoever resulting from loss of use, data or profits,
 * whether in an action of contract, negligence or other tortius action,
 * arising out of or in connection with the use or performance of this
 * software.
 *
 * Permission to use, copy, modify, and distribute this software for
 * any purpose and without fee is hereby granted, provided that the above
 * copyright notice and disclaimer, and also this permission notice, appear
 * in all copies.
 *
 */

#include <parmacs_internals.h>

/*
 * Data structures
 */
 
parmacs_data_t _parmacs_data = (parmacs_data_t)0;
parmacs_configuration _parmacs_configuration = {
  (char *)0,
  PARMACS_BUSY_WAIT_COUNT,
  PARMACS_MAX_BARRIER,
  PARMACS_MAX_GLOBAL,
  PARMACS_MAX_EVENT,
  PARMACS_MAX_LOCK,
  PARMACS_PAGE_SIZE,
  PARMACS_STACK_SIZE,
  PARMACS_SHARED_MEMORY,
  PARMACS_MEM_ALIGNMENT,
  PARMACS_TIMESTAMP
};
parmacs_configuration_t _parmacs_config = &_parmacs_configuration;

static void _parmacs_configure(unsigned long int);
static void _parmacs_cleanup(void);

/*
 * Initialize environment
 * Comments:
 *   This routine is supposed to be called from the main routine, at the
 *   very beginning of the execution.
 */
void
_parmacs_main_initenv(unsigned long int shm_size)
{
  register int ipc_id;
  register void *address;
  parmacs_semun sem_arg;
  register int i;
#ifdef PARMACS_NO_SCHED_YIELD
  int blocking_device[2];
#endif /* PARMACS_NO_SCHED_YIELD */
  
  /* Load configuration */
  _parmacs_configure(shm_size);
  
  /* Allocate memory for parmacs data */
  /* MAIN DATA */
  if ((ipc_id=shmget(IPC_PRIVATE, sizeof(parmacs_data), IPC_CREAT|0700))==-1){
    _parmacs_panic("parmacs: cannot allocate memory for parmacs data!");
  }
  if ((int)(address = shmat(ipc_id, (void *)0, 0)) == -1) {
    _parmacs_panic("parmacs: cannot attach memory for parmacs data!");
  }
  _parmacs_data = (parmacs_data_t)address;
  _parmacs_data->segment[PARMACS_MAIN_ID] = ipc_id;
  
  /* Initialize mutex locks */
  if ((ipc_id=semget(IPC_PRIVATE, PARMACS_MAX_ID, IPC_CREAT|0700)) == -1) {
    _parmacs_panic("parmacs: cannot allocate semaphore set for mutex!");
  }
  sem_arg.array = (u_short *)malloc(PARMACS_MAX_ID * sizeof(u_short));
  for(i = 0; i < PARMACS_MAX_ID; i++) sem_arg.array[i] = 1;
  if (semctl(ipc_id, 0, SETALL, sem_arg) == -1) {
    _parmacs_panic("parmacs: cannot initialize semaphore set for mutex!");
  }
  free(sem_arg.array);
  _parmacs_data->mutex = ipc_id;
  
  /* 
   * Initialize segments 
   */
   
  for (i = 1; i < PARMACS_MAX_ID; i++) _parmacs_data->segment[i] = -1;
  
  /* MEMORY */
  if ((ipc_id = shmget(IPC_PRIVATE,
                      _parmacs_config->shared_memory,
                      IPC_CREAT|0700)) ==-1){
    _parmacs_panic("parmacs: cannot allocate shared memory!");
  }
  if ((int)(address = shmat(ipc_id, (void *)0, 0)) == -1) {
    _parmacs_panic("parmacs: cannot attach shared memory!");
  }
  _parmacs_data->segment[PARMACS_MEMORY_ID] = ipc_id;
  _parmacs_data->shm_address = address;
  _parmacs_data->shm_pointer = address;
  _parmacs_data->shm_size    = _parmacs_config->shared_memory;
  _parmacs_data->shm_avail   = _parmacs_config->shared_memory;
  
  /* BARRIERS */
  if ((ipc_id = shmget(IPC_PRIVATE,
                      _parmacs_config->max_barriers * sizeof(parmacs_barrier),
                      IPC_CREAT|0700)) ==-1){
    _parmacs_panic("parmacs: cannot allocate memory for barriers!");
  }
  if ((int)(address = shmat(ipc_id, (void *)0, 0)) == -1) {
    _parmacs_panic("parmacs: cannot attach memory for barriers!");
  }
  _parmacs_data->segment[PARMACS_BARRIER_ID] = ipc_id;
  _parmacs_data->barrier = address;
  _parmacs_data->n_barriers = 0;
  
  /* GLOBAL SUBSCRIPTS */
  if ((ipc_id = shmget(IPC_PRIVATE,
                      _parmacs_config->max_globals * sizeof(parmacs_global),
                      IPC_CREAT|0700)) ==-1){
    _parmacs_panic("parmacs: cannot allocate memory for global subscripts!");
  }
  if ((int)(address = shmat(ipc_id, (void *)0, 0)) == -1) {
    _parmacs_panic("parmacs: cannot attach memory for global subscripts!");
  }
  _parmacs_data->segment[PARMACS_GLOBAL_ID] = ipc_id;
  _parmacs_data->global = address;
  _parmacs_data->n_globals = 0;
  
  /* EVENTS */
  if ((ipc_id = shmget(IPC_PRIVATE,
                      _parmacs_config->max_events * sizeof(parmacs_event),
                      IPC_CREAT|0700)) ==-1){
    _parmacs_panic("parmacs: cannot allocate memory for events!");
  }
  if ((int)(address = shmat(ipc_id, (void *)0, 0)) == -1) {
    _parmacs_panic("parmacs: cannot attach memory for events!");
  }
  _parmacs_data->segment[PARMACS_EVENT_ID] = ipc_id;
  _parmacs_data->event = address;
  _parmacs_data->n_events = 0;
  
  /* LOCKS */
  if ((ipc_id = shmget(IPC_PRIVATE,
                      _parmacs_config->max_locks * sizeof(parmacs_lock),
                      IPC_CREAT|0700)) ==-1){
    _parmacs_panic("parmacs: cannot allocate memory for locks!");
  }
  if ((int)(address = shmat(ipc_id, (void *)0, 0)) == -1) {
    _parmacs_panic("parmacs: cannot attach memory for locks!");
  }
  _parmacs_data->segment[PARMACS_LOCK_ID] = ipc_id;
  _parmacs_data->lock = address;
  _parmacs_data->n_locks = 0;
  
#ifdef PARMACS_NO_SCHED_YIELD  
  /*
   * Initialize blocking device to implement processor yields
   */
  if (pipe(blocking_device) < 0) {
    _parmacs_panic("parmacs: cannot create pipe for yielding!");
  }
  _parmacs_data->blocking_fd = blocking_device[0];
#endif /* PARMACS_NO_SCHED_YIELD */

  /*
   * Time stamp:
   */
  if (_parmacs_config->timestamp) {
    struct tms now;
    _parmacs_data->start_time = times(&now);
  }
}

void
_parmacs_main_end()
{
  /*
   * Time stamp:
   */
  if (_parmacs_config->timestamp) {
    struct tms now;
    _parmacs_data->end_time = times(&now);
    fprintf(stderr, "\nPARMACS TIME: start: %lu, end: %lu, elapsed: %lu.\n",
            _parmacs_data->start_time,
            _parmacs_data->end_time,
            _parmacs_data->end_time - _parmacs_data->start_time);
  }
  _parmacs_exit();
}

static void
_parmacs_configure(unsigned long int shm_size)
{
  extern char **environ;
  char **my_environment;
  char * var_name;

  _parmacs_configuration.shared_memory = shm_size;
  
  /* Read configuration from environment */
  
  for(my_environment = environ; *my_environment; my_environment++) {
    var_name = strtok(strdup(*my_environment), "=");
    if (!strcmp(var_name, "PARMACS_CONFIG_FILE")) {
      _parmacs_configuration.config_file = strdup(strtok(NULL, "\0"));
    } else if (!strcmp(var_name, "PARMACS_BUSY_WAIT_COUNT")) {
      _parmacs_configuration.busy_wait_count = atoi(strtok(NULL, "\0"));
    } else if (!strcmp(var_name, "PARMACS_MAX_BARRIER")) {
      _parmacs_configuration.max_barriers = atoi(strtok(NULL, "\0"));
    } else if (!strcmp(var_name, "PARMACS_MAX_GLOBAL")) {
      _parmacs_configuration.max_globals = atoi(strtok(NULL, "\0"));
    } else if (!strcmp(var_name, "PARMACS_MAX_EVENT")) {
      _parmacs_configuration.max_events = atoi(strtok(NULL, "\0"));
    } else if (!strcmp(var_name, "PARMACS_MAX_LOCK")) {
      _parmacs_configuration.max_locks = atoi(strtok(NULL, "\0"));
    } else if (!strcmp(var_name, "PARMACS_PAGE_SIZE")) {
      _parmacs_configuration.page_size = atoi(strtok(NULL, "\0"));
    } else if (!strcmp(var_name, "PARMACS_STACK_SIZE")) {
      _parmacs_configuration.stack_size = atoi(strtok(NULL, "\0"));
    } else if (!strcmp(var_name, "PARMACS_SHARED_MEMORY")) {
      char *endptr;
      _parmacs_configuration.shared_memory = 
        strtoul(strtok(NULL, "\0"), &endptr, 10);
    } else if (!strcmp(var_name, "PARMACS_MEM_ALIGNMENT")) {
      _parmacs_configuration.memory_alignment = atoi(strtok(NULL, "\0"));
    } else if (!strcmp(var_name, "PARMACS_TIMESTAMP")) {
      _parmacs_configuration.timestamp = atoi(strtok(NULL, "\0"));
    }
    free(var_name);  
  }
  
  /* if there is a configuration file, use it */
  if (_parmacs_configuration.config_file) {
    char buffer[512];
    FILE *fd;

    if (!(fd = fopen(_parmacs_configuration.config_file, "r"))) {
      _parmacs_panic("parmacs: cannot open configuration file!");
    }    
    while (fgets(buffer, 512, fd)) {
      var_name = strtok(buffer, " =\t\n\0");
      if (!strcmp(var_name, "PARMACS_BUSY_WAIT_COUNT")) {
        _parmacs_configuration.busy_wait_count = atoi(strtok(NULL, " =\t\n\0"));
      } else if (!strcmp(var_name, "PARMACS_MAX_BARRIER")) {
        _parmacs_configuration.max_barriers = atoi(strtok(NULL, " =\t\n\0"));
      } else if (!strcmp(var_name, "PARMACS_MAX_GLOBAL")) {
        _parmacs_configuration.max_globals = atoi(strtok(NULL, " =\t\n\0"));
      } else if (!strcmp(var_name, "PARMACS_MAX_EVENT")) {
        _parmacs_configuration.max_events = atoi(strtok(NULL, " =\t\n\0"));
      } else if (!strcmp(var_name, "PARMACS_MAX_LOCK")) {
        _parmacs_configuration.max_locks = atoi(strtok(NULL, " =\t\n\0"));
      } else if (!strcmp(var_name, "PARMACS_PAGE_SIZE")) {
        _parmacs_configuration.page_size = atoi(strtok(NULL, " =\t\n\0"));
      } else if (!strcmp(var_name, "PARMACS_STACK_SIZE")) {
        _parmacs_configuration.stack_size = atoi(strtok(NULL, " =\t\n\0"));
      } else if (!strcmp(var_name, "PARMACS_SHARED_MEMORY")) {
        char *endptr;
        _parmacs_configuration.shared_memory = 
          strtoul(strtok(NULL, " =\t\n\0"), &endptr, 10);
      } else if (!strcmp(var_name, "PARMACS_MEM_ALIGNMENT")) {
        _parmacs_configuration.memory_alignment=atoi(strtok(NULL, " =\t\n\0"));
      } else if (!strcmp(var_name, "PARMACS_TIMESTAMP")) {
        _parmacs_configuration.timestamp = atoi(strtok(NULL, " =\t\n\0"));
      }
    }
    fclose(fd);
  }
}

/*
 * Memory management
 * Comments:
 *   Memory management is extremely simple in this implementation. The
 *   allocation routine simply assigns consecutive regions within the
 *   shared memory area, till that area is full. Freeing memory is not
 *   implemented, so that memory can never be reused.
 *   The allocation routine returns an address which is properly aligned,
 *   or zero if there is not enough memory.
 */
 
void *
_parmacs_memory_alloc(int size)
{
  register void *address;
  register int vsize = size;
  register unsigned long int u_ptr;
  register int adjust;
  
  _parmacs_internal_lock(PARMACS_MEMORY_ID);
  u_ptr = (unsigned long int)_parmacs_data->shm_pointer;
  if ((adjust = u_ptr & (_parmacs_config->memory_alignment-1))) {
    u_ptr += _parmacs_config->memory_alignment - adjust;
    vsize += _parmacs_config->memory_alignment - adjust;
  }
  if ((adjust = size & (_parmacs_config->memory_alignment-1))) {
    size  += _parmacs_config->memory_alignment - adjust;
    vsize += _parmacs_config->memory_alignment - adjust;
  }
  if (vsize > _parmacs_data->shm_avail) {
    _parmacs_internal_unlock(PARMACS_MEMORY_ID);
    return (void *)0;
  }
  address = (void *)u_ptr;
  _parmacs_data->shm_pointer = (void *)(u_ptr + size);
  _parmacs_data->shm_avail -= vsize;
  _parmacs_internal_unlock(PARMACS_MEMORY_ID);
  return (void *)address;
}

void
_parmacs_memory_free(void *address)
{
  /* not implemented */
}

void *
_parmacs_local_alloc(int size)
{
  void *address;
  _parmacs_internal_lock(PARMACS_MEMORY_ID);
  address = (void *)malloc(size);
  _parmacs_internal_unlock(PARMACS_MEMORY_ID);
  return address;
}

void
_parmacs_local_free(void *address)
{
  _parmacs_internal_lock(PARMACS_MEMORY_ID);
  free(address);
  _parmacs_internal_unlock(PARMACS_MEMORY_ID);
}

/*
 * Process creation and joining
 */
 
void
_parmacs_create(void (*routine)())
{
  switch(fork()) {
  case -1:
    _parmacs_panic("parmacs: cannot create new process!");
  case 0:
    (*routine)();
    exit(0);
  default:
    break;
  }
}

void
_parmacs_wait_for_end(int amount)
{
  register int i;
  int status;
  
  for(i = 0; i < amount; i++) {
    if (wait(&status) == -1 && errno == EINTR) i--;
  }
}

parmacs_pid
_parmacs_get_pid()
{
  return ((parmacs_pid)getpid());
}

/*
 * Scheduling, clean up and exit routines
 */
 
static void 
_parmacs_cleanup()
{
  register int i;
  parmacs_semun sem_arg;
  struct shmid_ds *shm_arg = (struct shmid_ds  *)0;
  
  sem_arg.val = 0;
  if (_parmacs_data) {
    for (i=0; i < _parmacs_data->n_barriers; i++) {
      semctl(_parmacs_data->barrier[i].s, 0, IPC_RMID, sem_arg);
    }
    for (i=0; i < _parmacs_data->n_globals; i++) {
      semctl(_parmacs_data->global[i].s, 0, IPC_RMID, sem_arg);
    }
    for (i=0; i < _parmacs_data->n_events; i++) {
      semctl(_parmacs_data->event[i].s, 0, IPC_RMID, sem_arg);
    }
    for (i=0; i < _parmacs_data->n_locks; i++) {
      semctl(_parmacs_data->lock[i].s, 0, IPC_RMID, sem_arg);
    }
    semctl(_parmacs_data->mutex, 0, IPC_RMID, sem_arg);
    
    for (i=1; i < PARMACS_MAX_ID; i++) {
      if (_parmacs_data->segment[i] != -1) {
        shmctl(_parmacs_data->segment[i], IPC_RMID, shm_arg);
      }
    }
    shmctl(_parmacs_data->segment[0], IPC_RMID, shm_arg);
  }
}
 
void 
_parmacs_panic(char *msg)
{
  if (msg) fprintf(stderr, "%s\n", msg);
  _parmacs_cleanup();
  kill(0, SIGKILL);
  exit(-1);
}

void 
_parmacs_exit()
{
  _parmacs_cleanup();
  exit(0);
}

void 
_parmacs_yield()
{
#ifndef PARMACS_NO_SCHED_YIELD
  sched_yield();
#else /* PARMACS_NO_SCHED_YIELD */
  struct timeval t;
  fd_set fds1, fds2, fds3;
  t.tv_sec  = 0;
  t.tv_usec = 1;
  FD_ZERO(&fds1);
  FD_ZERO(&fds2);
  FD_ZERO(&fds3);
  FD_SET(_parmacs_data->blocking_fd, &fds1);
  select(1, &fds1, &fds2, &fds3, &t);
#endif /* PARMACS_NO_SCHED_YIELD */
}

/*
 * Clock
 */
parmacs_timestamp
_parmacs_clock()
{
  struct tms now;
  return (parmacs_timestamp)times(&now);
}

