/*
 * File: sysv/parmacs_sync.c
 * Description: implementation of synchronization macros of PARMACS
 * Comments: 
 * Copyright: Department of Computer Architecture - Polytechnic University of
 *   Catalonia (UPC)
 * Created: 1999/06 ernest@ac.upc.es (E.Artiaga)
 * Contact: gsomk@ac.upc.es
 * Last update:
 * History:
 */
/*
 * Disclaimer and copying policy:
 *
 * The Polytechnic University of Catalonia (UPC) and the authors disclaim
 * all warranties with regard to this software, including all implied
 * warranties of merchantability and fitness.
 * In no event shall the Polytechnic University of Catalonia (UPC) or the
 * authors be liable for any special, indirect or consequential damages or
 * any damages whatsoever resulting from loss of use, data or profits,
 * whether in an action of contract, negligence or other tortius action,
 * arising out of or in connection with the use or performance of this
 * software.
 *
 * Permission to use, copy, modify, and distribute this software for
 * any purpose and without fee is hereby granted, provided that the above
 * copyright notice and disclaimer, and also this permission notice, appear
 * in all copies.
 *
 */

#include <parmacs_internals.h>


/*
 * Barriers
 */

void
_parmacs_barinit(parmacs_barrier_t *barrier)
{
  parmacs_semun sem_arg;
  
  /* get a free barrier entry */
  _parmacs_internal_lock(PARMACS_BARRIER_ID);
  if (_parmacs_data->n_barriers >= _parmacs_config->max_barriers) {
    _parmacs_internal_unlock(PARMACS_BARRIER_ID);
    _parmacs_panic("parmacs: too many barriers!");
  }
  *barrier = &_parmacs_data->barrier[_parmacs_data->n_barriers++];
  _parmacs_internal_unlock(PARMACS_BARRIER_ID);
  
  /* now, initialize the new barrier */
  if (((*barrier)->s = semget(IPC_PRIVATE, 2, IPC_CREAT|0700)) == -1) {
    _parmacs_panic("parmacs: cannot allocate semaphore for barrier!");
  }
  sem_arg.val = 1;
  if (semctl((*barrier)->s, 0, SETVAL, sem_arg) == -1) {
    _parmacs_panic("parmacs: cannot initialize mutex semaphore for barrier!");
  }
  sem_arg.val = 0;
  if (semctl((*barrier)->s, 1, SETVAL, sem_arg) == -1) {
    _parmacs_panic(
      "parmacs: cannot initialize blocking semaphore for barrier!");
  }
  (*barrier)->n = 0;
}
 
void 
_parmacs_barrier(parmacs_barrier_t barrier, int procs)
{
  parmacs_semun sem_arg;
  struct sembuf operation;
  register int countdown;
  register int spin = _parmacs_config->busy_wait_count;
  register int wakeup_or_block;
  
  /* First, wait for everybody to go out from the previous barrier call */
  countdown = spin;
  while(barrier->n < 0) {
    if (countdown >= 0) {
      if (!countdown) {
        _parmacs_yield();
        countdown = spin;
      } else {
        countdown --;
      }
    }
  }
  
  /* Now enter the barrier */
  _parmacs_sem_lock(barrier->s, 0);
  barrier->n++;
  if (barrier->n == 1) {
    /* This is the first process - setup the barrier */
    sem_arg.val = procs;
    semctl(barrier->s, 1, SETVAL, sem_arg);
  }
  if (barrier->n == procs) {
    /* This is the last process - wakeup everybody */
    wakeup_or_block = -procs;
    /* close barrier until everybody is out */
    barrier->n = -procs;
  } else {
    wakeup_or_block = 0;
  }
  _parmacs_sem_unlock(barrier->s, 0);
  
  /* Actually block or wakeup everybody */   
  operation.sem_num = 1;
  operation.sem_op  = wakeup_or_block;
  operation.sem_flg = 0;
  semop(barrier->s, &operation, 1);
  
  /* say this process got out the barrier */
  _parmacs_sem_lock(barrier->s, 0);
  barrier->n++;
  _parmacs_sem_unlock(barrier->s, 0);
}

/*
 * Global subscripts
 */
 
void
_parmacs_gsinit(parmacs_global_t *global)
{
  parmacs_semun sem_arg;
  
  /* get a free global subscript entry */
  _parmacs_internal_lock(PARMACS_GLOBAL_ID);
  if (_parmacs_data->n_globals >= _parmacs_config->max_globals) {
    _parmacs_internal_unlock(PARMACS_GLOBAL_ID);
    _parmacs_panic("parmacs: too many global subscripts!");
  }
  *global = &_parmacs_data->global[_parmacs_data->n_globals++];
  _parmacs_internal_unlock(PARMACS_GLOBAL_ID);
  
  /* now, initialize the new global subscript */
  if (((*global)->s = semget(IPC_PRIVATE, 2, IPC_CREAT|0700)) == -1) {
    _parmacs_panic("parmacs: cannot allocate semaphore for global subscript!");
  }
  sem_arg.val = 1;
  if (semctl((*global)->s, 0, SETVAL, sem_arg) == -1) {
    _parmacs_panic(
      "parmacs: cannot initialize mutex semaphore for global subscript!");
  }
  sem_arg.val = 0;
  if (semctl((*global)->s, 1, SETVAL, sem_arg) == -1) {
    _parmacs_panic(
      "parmacs: cannot initialize blocking semaphore for global subscript!");
  }
  (*global)->n = 0;
  (*global)->i = 0;
}

void
_parmacs_getsub(parmacs_global_t global, int *next, int last, int procs)
{
  parmacs_semun sem_arg;
  struct sembuf operation;
  register int countdown;
  register int spin = _parmacs_config->busy_wait_count;
  register int wakeup_or_block;
  
  /* First, wait for global subscript to be ready */
  countdown = spin;
  while(global->n < 0) {
    if (countdown >= 0) {
      if (!countdown) {
        _parmacs_yield();
        countdown = spin;
      } else {
        countdown --;
      }
    }
  }
  
  /* Now get next subscript */
  _parmacs_sem_lock(global->s, 0);
  if (global->i <= last) {
    *next = global->i++;
  } else {
    *next = -1;
    /* the loop is finished: block until everybody is done */
    global->n++;
    if (global->n == 1) {
      /* This is the first process - setup the barrier */
      sem_arg.val = procs;
      semctl(global->s, 1, SETVAL, sem_arg);
    }
    if (global->n == procs) {
      /* This is the last process - wakeup everybody */
      wakeup_or_block = -procs;
      /* close global subscript's barrier until everybody is out */
      global->n = -procs;
      /* reset subscript */
      global->i = 0;
    } else {
      wakeup_or_block = 0;
    }
  }
  _parmacs_sem_unlock(global->s, 0);
  
  if (*next == -1) {
    /* Actually block or wakeup everybody */   
    operation.sem_num = 1;
    operation.sem_op  = wakeup_or_block;
    operation.sem_flg = 0;
    semop(global->s, &operation, 1);
  
    /* say this process got out the barrier */
    _parmacs_sem_lock(global->s, 0);
    global->n++;
    _parmacs_sem_unlock(global->s, 0);
  }
}

/*
 * Locks
 */

void 
_parmacs_lockinit(parmacs_lock_t *lock)
{
  parmacs_semun sem_arg;
  
  /* get a free lock entry */
  _parmacs_internal_lock(PARMACS_LOCK_ID);
  if (_parmacs_data->n_locks >= _parmacs_config->max_locks) {
    _parmacs_internal_unlock(PARMACS_LOCK_ID);
    _parmacs_panic("parmacs: too many locks!");
  }
  *lock = &_parmacs_data->lock[_parmacs_data->n_locks++];
  _parmacs_internal_unlock(PARMACS_LOCK_ID);
  
  /* now, initialize the new lock */
  if (((*lock)->s = semget(IPC_PRIVATE, 1, IPC_CREAT|0700)) == -1) {
    _parmacs_panic("parmacs: cannot allocate semaphore for lock!");
  }
  sem_arg.val = 1;
  if (semctl((*lock)->s, 0, SETVAL, sem_arg) == -1) {
    _parmacs_panic("parmacs: cannot initialize semaphore for lock!");
  }
}

void
_parmacs_lock(parmacs_lock_t lock)
{
  _parmacs_sem_lock(lock->s, 0);
}

void
_parmacs_unlock(parmacs_lock_t lock)
{
  _parmacs_sem_unlock(lock->s, 0);
}

void
_parmacs_alockinit(parmacs_lock_t *lock, int amount)
{
  parmacs_semun sem_arg;
  register int i;
  
  /* get a free lock entry */
  _parmacs_internal_lock(PARMACS_LOCK_ID);
  if (_parmacs_data->n_locks >= _parmacs_config->max_locks) {
    _parmacs_internal_unlock(PARMACS_LOCK_ID);
    _parmacs_panic("parmacs: too many locks!");
  }
  *lock = &_parmacs_data->lock[_parmacs_data->n_locks++];
  _parmacs_internal_unlock(PARMACS_LOCK_ID);
  
  /* now, initialize the new lock */
  if (((*lock)->s = semget(IPC_PRIVATE, amount, IPC_CREAT|0700)) == -1) {
    _parmacs_panic("parmacs: cannot allocate semaphore set for lock array!");
  }
  
  if (!(sem_arg.array=(u_short*)_parmacs_local_alloc(amount*sizeof(u_short)))){
    _parmacs_panic("parmacs: cannot allocate memory to initialize lock array!");
  }
  for(i=0; i < amount; i++) {
    sem_arg.array[i] = 1;
  }
  
  if (semctl((*lock)->s, 0, SETALL, sem_arg) == -1) {
    _parmacs_panic("parmacs: cannot initialize semaphore for lock array!");
  }
  
  _parmacs_local_free(sem_arg.array);
}

void
_parmacs_alock(parmacs_lock_t lock, int idx)
{
  _parmacs_sem_lock(lock->s, idx);
}

void
_parmacs_aulock(parmacs_lock_t lock, int idx)
{
  _parmacs_sem_unlock(lock->s, idx);
}

/*
 * Events
 */
 
void
_parmacs_pauseinit(parmacs_event_t *event, int amount)
{
  register int i;
  parmacs_semun sem_arg;
  u_short array[3];
  
  /* get the free event entries */
  _parmacs_internal_lock(PARMACS_EVENT_ID);
  if ((_parmacs_data->n_events + amount -1) >= _parmacs_config->max_events) {
    _parmacs_internal_unlock(PARMACS_EVENT_ID);
    _parmacs_panic("parmacs: too many events!");
  }
  *event = &_parmacs_data->event[_parmacs_data->n_events];
  _parmacs_data->n_events += amount;
  _parmacs_internal_unlock(PARMACS_EVENT_ID);
  
  /* initialize the event entries */
  sem_arg.array = array;
  for(i=0; i < amount; i++) {
    if (((*event)[i].s = semget(IPC_PRIVATE, 3, IPC_CREAT|0700)) == -1) {
      _parmacs_panic("parmacs: cannot allocate semaphore for event!");
    }
    sem_arg.array[0] = 1;
    sem_arg.array[1] = 1;
    sem_arg.array[2] = 0;
    if (semctl((*event)[i].s, 0, SETALL, sem_arg) == -1) {
      _parmacs_panic("parmacs: cannot initialize semaphore set for event!");
    }
  }
}

void
_parmacs_clearpause(parmacs_event_t event, int idx)
{
  parmacs_semun sem_arg;
  
  _parmacs_sem_lock(event[idx].s, 0);
  sem_arg.val = 1;
  semctl(event[idx].s, 1, SETVAL, sem_arg);
  sem_arg.val = 0;
  semctl(event[idx].s, 2, SETVAL, sem_arg);
  _parmacs_sem_unlock(event[idx].s, 0);
}

void
_parmacs_setpause(parmacs_event_t event, int idx)
{
  parmacs_semun sem_arg;
  
  _parmacs_sem_lock(event[idx].s, 0);
  sem_arg.val = 0;
  semctl(event[idx].s, 1, SETVAL, sem_arg);
  sem_arg.val = 1;
  semctl(event[idx].s, 2, SETVAL, sem_arg);
  _parmacs_sem_unlock(event[idx].s, 0);
}

void
_parmacs_event(parmacs_event_t event, int idx)
{
  struct sembuf operation[2];
  parmacs_semun sem_arg;
  
  operation[0].sem_num = 0;
  operation[0].sem_op  = -1;
  operation[0].sem_flg = 0;
  
  operation[1].sem_num = 1;
  operation[1].sem_op  = -1;
  operation[1].sem_flg = 0;
  
  sem_arg.val = 1;
  
  semop(event[idx].s, operation, 2);
  semctl(event[idx].s, 2, SETVAL, sem_arg);

  operation[0].sem_num = 0;
  operation[0].sem_op  = 1;
  operation[0].sem_flg = 0;
  
  semop(event[idx].s, operation, 1);
}

void
_parmacs_waitpause(parmacs_event_t event, int idx)
{
  struct sembuf operation;
  
  operation.sem_num = 1;
  operation.sem_op  = 0;
  operation.sem_flg = 0;
  semop(event[idx].s, &operation, 1);
}

void
_parmacs_pause(parmacs_event_t event, int idx)
{
  struct sembuf operation[3];
  parmacs_semun sem_arg;
  
  operation[0].sem_num =  0;
  operation[0].sem_op  = -1;
  operation[0].sem_flg =  0;
  
  operation[1].sem_num =  1;
  operation[1].sem_op  =  0;
  operation[1].sem_flg =  0;
  
  operation[2].sem_num =  2;
  operation[2].sem_op  = -1;
  operation[2].sem_flg =  0;
 
  semop(event[idx].s, operation, 3);
  sem_arg.val = 0;
  semctl(event[idx].s, 2, SETVAL, sem_arg);
  sem_arg.val = 1;
  semctl(event[idx].s, 1, SETVAL, sem_arg);

  operation[0].sem_num =  0;
  operation[0].sem_op  =  1;
  operation[0].sem_flg =  0;
  semop(event[idx].s, operation, 1);
}  
  
  
